# -*- coding: utf-8 -*-
from __future__ import annotations

import secrets
import string

__all__ = [
    'make_random_number_string',
    'make_random_alphabet_string',
    'make_random_ascii_string',
    'make_salt'
]


def make_random_number_string(nchars: int) -> str:
    return ''.join(secrets.choice(string.digits) for _ in range(nchars))


def make_random_alphabet_string(nchars: int) -> str:
    return ''.join(secrets.choice(string.ascii_letters) for _ in range(nchars))


def make_random_ascii_string(nchars: int) -> str:
    return ''.join(secrets.choice(string.ascii_letters + string.digits) for _ in range(nchars))


def make_salt(nbytes: int) -> bytes:
    return secrets.token_bytes(nbytes)



# -*- coding: utf-8 -*-
from __future__ import annotations

import re
import warnings
from base64 import b64decode, b64encode
from dataclasses import dataclass
from pathlib import Path
from typing import Callable, IO, Iterable, Literal, NamedTuple, overload

from .qmcdataciphers import HardenedRC4, Mask128
from .qmckeyciphers import QMCv2KeyEncryptV1, QMCv2KeyEncryptV2
from ..exceptions import CrypterCreatingError
from ..keyutils import make_random_ascii_string, make_salt
from ..miscutils import proberfuncfactory
from ..prototypes import EncryptedBytesIO
from ..typedefs import BytesLike, FilePath, IntegerLike
from ..typeutils import isfilepath, tobytes, verify_fileobj
from ..warns import CrypterSavingWarning

warnings.filterwarnings(action='default', category=CrypterSavingWarning, module=__name__)
warnings.filterwarnings(action='default', category=DeprecationWarning, module=__name__)

__all__ = [
    'probe_qmc',
    'probe_qmcv1',
    'probe_qmcv2',
    'QMCv1',
    'QMCv2',
    'QMCv2QTag',
    'QMCv2STag',
    'QMCv1FileInfo',
    'QMCv2FileInfo'
]

QMCV1_FILENAME_PATTERN = re.compile(r'^.*\.qmc([0-9]|flac|ogg|ra)$', flags=re.IGNORECASE)
QMCv2_FILENAME_PATTERN = re.compile(r'^.*\.m(flac|gg)[0-9a-zA-Z]?$', flags=re.IGNORECASE)


@dataclass
class QMCv2QTag:
    """解析、存储和生成 QMCv2 文件末尾的 QTag 数据（不包括主密钥）。

    可以按照操作数据类（``dataclass``）实例的方式操作本类的实例。
    """
    song_id: int = 0
    """QTag 数据的第二部分，为 QTag 数据所在文件中被加密的歌曲在 QQ 音乐的 ID。"""
    unknown: int = 2
    """QTag 数据的第三部分，含义未知，在已知所有样本中都为 2。"""

    @classmethod
    def load(cls, qtag_serialized: BytesLike, /):
        """解析一段 QTag 数据，返回加密后的主密钥和一个 QMCv2QTag 对象。"""
        qtag_serialized = tobytes(qtag_serialized)
        qtag_serialized_splitted = qtag_serialized.split(b',')
        if len(qtag_serialized_splitted) != 3:
            raise ValueError('invalid QMCv2 QTag data: the counts of splitted segments '
                             f'should be equal to 3, not {len(qtag_serialized_splitted)}'
                             )
        master_key_encrypted_b64encoded = qtag_serialized_splitted[0]
        song_id: int = int(qtag_serialized_splitted[1])
        unknown: int = int(qtag_serialized_splitted[2])

        return master_key_encrypted_b64encoded, cls(song_id=song_id, unknown=unknown)

    def dump(self, master_key_encrypted_b64encoded: BytesLike, /) -> bytes:
        """根据当前 QMCv2QTag 对象生成并返回一段 QTag 数据，需要已加密的主密钥。"""
        return b','.join(
            [
                tobytes(master_key_encrypted_b64encoded),
                str(self.song_id).encode('ascii'),
                str(self.unknown).encode('ascii')
            ]
        )


@dataclass
class QMCv2STag:
    """解析、存储和重建 QMCv2 文件末尾的 STag 数据。

    可以按照操作数据类（``dataclass``）实例的方式操作本类的实例。
    """
    song_id: int = 0
    """STag 数据的第一部分，为 QTag 数据所在文件中被加密的歌曲在 QQ 音乐的 ID。"""
    unknown: int = 2
    """STag 数据的第二部分，含义未知，在已知所有样本中都为 2。"""
    song_mid: str = '0' * 14
    """STag 数据的第三部分，为 QTag 数据所在文件中被加密的歌曲在 QQ 音乐的媒体 ID（MId）。"""

    @classmethod
    def load(cls, stag_serialized: BytesLike, /):
        """解析一段 STag 数据，返回一个 QMCv2STag 对象。"""
        stag_serialized = tobytes(stag_serialized)
        stag_serialized_splitted = stag_serialized.split(b',')
        if len(stag_serialized_splitted) != 3:
            raise ValueError('invalid QMCv2 STag data: the counts of splitted segments '
                             f'should be equal to 3, not {len(stag_serialized_splitted)}'
                             )
        song_id: int = int(stag_serialized_splitted[0])
        unknown: int = int(stag_serialized_splitted[1])
        song_mid: str = str(stag_serialized_splitted[2], encoding='ascii')

        return cls(song_id=song_id, unknown=unknown, song_mid=song_mid)

    def dump(self) -> bytes:
        """根据当前 QMCv2STag 对象生成并返回一段 STag 数据。"""
        return b','.join(
            [
                str(self.song_id).encode('ascii'),
                str(self.unknown).encode('ascii'),
                str(self.song_mid).encode('ascii')
            ]
        )


class QMCv1FileInfo(NamedTuple):
    """用于存储 QMCv1 文件的信息。"""
    cipher_data_offset: int
    """加密数据在文件中开始的位置。"""
    cipher_data_len: int
    """加密数据在文件中的长度。"""
    opener: Callable[[tuple[FilePath | IO[bytes], QMCv1FileInfo] | FilePath | IO[bytes], ...], QMCv1]
    """打开文件的方式，为一个可调对象，其会返回一个加密文件对象。"""
    opener_kwargs_required: tuple[str, ...]
    """通过 ``opener`` 打开文件时，所必需的关键字参数的名称。"""
    opener_kwargs_optional: tuple[str, ...]
    """通过 ``opener`` 打开文件时，可选的关键字参数的名称。
    
    此属性仅储存可能会影响 ``opener`` 行为的可选关键字参数；
    对 ``opener`` 行为没有影响的可选关键字参数不会出现在此属性中。
    """


class QMCv2FileInfo(NamedTuple):
    """用于存储 QMCv2 文件的信息。"""
    cipher_ctor: Callable[[...], HardenedRC4] | Callable[[...], Mask128] | None
    """Cipher 的构造函数，接受一个必需参数（密钥），返回一个用于解密数据的
    Cipher 对象。通常就是 Cipher 类自身。
    
    对于文件尾部嵌有 STag 数据，或者无任何尾部数据的 QMCv2 文件，由于其没有内嵌密钥，此属性为 ``None``。
    """
    cipher_data_offset: int
    """加密数据在文件中开始的位置。"""
    cipher_data_len: int
    """加密数据在文件中的长度。"""
    master_key_encrypted: bytes
    """受加密保护的主密钥。
    
    对于文件尾部嵌有 STag 数据，或者无任何尾部数据的 QMCv2 文件，由于其没有内嵌密钥，此属性为 ``None``。
    """
    master_key_encryption_ver: int | None
    """主密钥的加密保护版本，通常为 1 或 2，也仅支持这两个版本。
    
    对于文件尾部嵌有 STag 数据，或者无任何尾部数据的 QMCv2 文件，由于其没有内嵌密钥，此属性为 ``None``。
    """
    extra_info: QMCv2QTag | QMCv2STag | None
    """文件尾部嵌有的除主密钥之外的额外数据，经过解析后的结果。
    如果有，为 ``QMCv2QTag`` 或 ``QMCv2STag`` 对象；如果没有，则为 ``None``。
    """
    opener: Callable[[tuple[FilePath | IO[bytes], QMCv2FileInfo] | FilePath | IO[bytes], ...], QMCv2]
    """打开文件的方式，为一个可调对象，其会返回一个加密文件对象。"""
    opener_kwargs_required: tuple[str, ...]
    """通过 ``opener`` 打开文件时，所必需的关键字参数的名称。"""
    opener_kwargs_optional: tuple[str, ...]
    """通过 ``opener`` 打开文件时，可选的关键字参数的名称。
    
    此属性仅储存可能会影响 ``opener`` 行为的可选关键字参数；
    对 ``opener`` 行为没有影响的可选关键字参数不会出现在此属性中。
    """


def _guess_cipher_ctor(master_key: BytesLike, /,
                       is_encrypted: bool = True
                       ) -> Callable[[...], HardenedRC4] | Callable[[...], Mask128] | None:
    if is_encrypted:
        expected_keylen_mask128 = (272, 392)
        expected_keylen_hardened_rc4 = (528, 736)
    else:
        expected_keylen_mask128 = (256, 256)
        expected_keylen_hardened_rc4 = (512, 512)

    master_key = tobytes(master_key)
    if len(master_key) in expected_keylen_mask128:
        return Mask128.from_qmcv2_key256
    elif len(master_key) in expected_keylen_hardened_rc4:
        return HardenedRC4
    elif len(master_key) == 128 and not is_encrypted:
        return Mask128


@overload
def probe_qmcv1(filething: FilePath | IO[bytes], /,
                is_qmcv1: bool = False
                ) -> tuple[Path | IO[bytes], QMCv1FileInfo | None]:
    pass


@proberfuncfactory
def probe_qmcv1(filething, /, is_qmcv1=False):
    r"""探测源文件 ``filething`` 是否为一个 QMCv1 文件。

    返回一个 2 个元素长度的元组：

    - 第一个元素为 ``filething``；
    - 如果 ``filething`` 是 QMCv1 文件，那么第二个元素为一个 ``QMCv1FileInfo`` 对象；
    - 否则为 ``None``。

    目前无法在没有密钥的情况下通过文件结构识别 QMCv1 文件，
    因此本方法通过文件名模式判断是否为 QMCv1 文件。
    只要文件名匹配下列正则表达式模式（不区分大小写），本方法就会将此文件视为一个 QMCv1 文件：

    ``^.*\.qmc([0-9]|flac|ogg|ra)$``

    对于不匹配以上正则表达式的文件名，如果参数 ``is_qmcv1=True``，
    本方法会跳过探测过程，认为此文件是一个 QMCv1 文件，并直接返回结果。

    本方法的返回值可以用于 ``QMCv1.open()`` 的第一个位置参数。

    本方法不适用于 QMCv2 文件的探测。

    Args:
        filething: 源文件的路径或文件对象
        is_qmcv1: 跳过探测过程，认为源文件是一个 QMCv1 文件
    Returns:
        一个 2 个元素长度的元组：第一个元素为 filething；如果
        filething 是 QMCv1 文件，那么第二个元素为一个 QMCv1FileInfo 对象；否则为 None。
    """

    filename = getattr(filething, 'name', None)
    if filename is None and not is_qmcv1:
        return
    filepath = Path(filename)
    if QMCV1_FILENAME_PATTERN.fullmatch(filepath.name) or is_qmcv1:
        return QMCv1FileInfo(
            cipher_data_offset=0,
            cipher_data_len=filething.seek(0, 2),
            opener=QMCv1.open,
            opener_kwargs_required=('mask',),
            opener_kwargs_optional=()
        )


@overload
def probe_qmcv2(filething: FilePath | IO[bytes], /) -> tuple[Path | IO[bytes], QMCv2FileInfo | None]:
    pass


@proberfuncfactory
def probe_qmcv2(filething, /):
    """探测源文件 ``filething`` 是否为一个 QMCv2 文件。

    返回一个 2 个元素长度的元组：

    - 第一个元素为 ``filething``；
    - 如果 ``filething`` 是 QMCv2 文件，那么第二个元素为一个 ``QMCv2FileInfo`` 对象；
    - 否则为 ``None``。

    本方法的返回值可以用于 ``QMCv2.open()`` 的第一个位置参数。

    本方法不适用于 QMCv1 文件的探测。

    Args:
        filething: 源文件的路径或文件对象
    Returns:
        一个 2 个元素长度的元组：第一个元素为 filething；如果
        filething 是 QMCv2 文件，那么第二个元素为一个 QMCv2FileInfo 对象；否则为 None。
    """
    opener_kwargs_required = []
    opener_kwargs_optional = []

    total_size = filething.seek(-4, 2) + 4
    tail_data = filething.read(4)

    if tail_data == b'STag':
        filething.seek(-8, 2)
        tag_serialized_len = int.from_bytes(filething.read(4), 'big')
        if tag_serialized_len > (total_size - 8):
            return
        cipher_data_len = filething.seek(-(tag_serialized_len + 8), 2)
        extra_info = QMCv2STag.load(filething.read(tag_serialized_len))

        cipher_ctor = None
        master_key_encrypted = None
        master_key_encryption_ver = None

        opener_kwargs_required.append('master_key')
    elif tail_data == b'QTag':
        opener_kwargs_required.append('core_key')
        opener_kwargs_optional.append('master_key')

        filething.seek(-8, 2)
        tag_serialized_len = int.from_bytes(filething.read(4), 'big')
        if tag_serialized_len > (total_size - 8):
            return
        cipher_data_len = filething.seek(-(tag_serialized_len + 8), 2)
        master_key_encrypted_b64encoded, extra_info = QMCv2QTag.load(filething.read(tag_serialized_len))
        try:
            master_key_encrypted_b64encoded.decode('ascii')
        except UnicodeDecodeError:
            return
        master_key_encrypted_b64decoded = b64decode(master_key_encrypted_b64encoded)
        if master_key_encrypted_b64decoded.startswith(b'QQMusic EncV2,Key:'):
            master_key_encrypted = master_key_encrypted_b64decoded[18:]
            master_key_encryption_ver = 2

            opener_kwargs_required.append('garble_keys')
        else:
            master_key_encrypted = master_key_encrypted_b64decoded
            master_key_encryption_ver = 1
        cipher_ctor = _guess_cipher_ctor(master_key_encrypted)
    else:
        opener_kwargs_required.append('core_key')
        opener_kwargs_optional.append('master_key')

        extra_info = None
        master_key_encrypted_b64encoded_len = int.from_bytes(tail_data, 'little')
        if master_key_encrypted_b64encoded_len > total_size - 4:
            return
        cipher_data_len = filething.seek(-(master_key_encrypted_b64encoded_len + 4), 2)
        master_key_encrypted_b64encoded = filething.read(master_key_encrypted_b64encoded_len)
        try:
            master_key_encrypted_b64encoded.decode('ascii')
        except UnicodeDecodeError:
            return
        master_key_encrypted_b64decoded = b64decode(master_key_encrypted_b64encoded)
        if master_key_encrypted_b64decoded.startswith(b'QQMusic EncV2,Key:'):
            master_key_encrypted = master_key_encrypted_b64decoded[18:]
            master_key_encryption_ver = 2

            opener_kwargs_required.append('garble_keys')
        else:
            master_key_encrypted = master_key_encrypted_b64decoded
            master_key_encryption_ver = 1
        cipher_ctor = _guess_cipher_ctor(master_key_encrypted)

    return QMCv2FileInfo(cipher_ctor=cipher_ctor,
                         cipher_data_offset=0,
                         cipher_data_len=cipher_data_len,
                         master_key_encrypted=master_key_encrypted,
                         master_key_encryption_ver=master_key_encryption_ver,
                         extra_info=extra_info,
                         opener=QMCv2.open,
                         opener_kwargs_required=tuple(opener_kwargs_required),
                         opener_kwargs_optional=tuple(opener_kwargs_optional)
                         )


def probe_qmc(
        filething: FilePath | IO[bytes], /,
        is_qmcv1: bool = False
) -> tuple[Path | IO[bytes], QMCv1FileInfo | None] | tuple[Path | IO[bytes], QMCv2FileInfo | None]:
    """探测源文件 ``filething`` 是否为一个 QMCv1 或 QMCv2 文件。

    返回一个 2 个元素长度的元组：

    - 第一个元素为 ``filething``；
    - 如果 ``filething`` 是 QMCv1 文件，那么第二个元素为一个 ``QMCv1FileInfo`` 对象；
    - 如果 ``filething`` 是 QMCv2 文件，那么第二个元素为一个 ``QMCv2FileInfo`` 对象；
    - 如果都不是，则为 ``None``。

    本方法的返回值可以用于 ``QMCv1.open()`` 和 ``QMCv2.open()`` 的第一个位置参数。

    Args:
        filething: 源文件的路径或文件对象
        is_qmcv1: （针对 QMCv1 文件）跳过探测过程，认为源文件是一个 QMCv1 文件；QMCv2 文件不受此参数影响
    Returns:
        一个 2 个元素长度的元组：第一个元素为 filething；如果
        filething 是 QMCv1 文件，那么第二个元素为一个 QMCv1FileInfo 对象；如果
        filething 是 QMCv2 文件，那么第二个元素为一个 QMCv2FileInfo 对象；否则为 None。
    """
    fthing, fileinfo = probe_qmcv2(filething)
    if fileinfo:
        return fthing, fileinfo
    return probe_qmcv1(filething, is_qmcv1=is_qmcv1)


class QMCv1(EncryptedBytesIO):
    """基于 BytesIO 的 QMCv1 透明加密二进制流。

    所有读写相关方法都会经过透明加密层处理：
    读取时，返回解密后的数据；写入时，向缓冲区写入加密后的数据。

    调用读写相关方法时，附加参数 ``nocryptlayer=True``
    可绕过透明加密层，访问缓冲区内的原始加密数据。

    如果你要新建一个 QMCv1 对象，不要直接调用 ``__init__()``，而是使用构造器方法
    ``QMCv1.new()`` 和 ``QMCv1.open()`` 新建或打开已有 QMCv1 文件，
    使用已有 QMCv1 对象的 ``save()`` 方法将其保存到文件。

    使用示例：

    - 新建一个 QMCv1 对象以便编辑：
    >>> qmcv1file = QMCv1.new()  # 添加 mask 参数以便自定义密钥，否则使用随机生成的密钥
    >>> qmcv1file
    <libtakiyasha.qmc.QMCv1 at 0x7ff6820d9e90, cipher <libtakiyasha.qmc.qmcdataciphers.Mask128 object at 0x7ff682178c10>>
    >>>

    - 获取使用的主密钥：
    >>> qmcv1file.master_key  # 此处的密钥是随机生成的
    b'}t4\\x87-\\xcd\\x88\\x93Ro\\x12g1Q\\xc4O\\xb3...'
    >>>

    - 访问内部的 Cipher 对象：
    >>> qmcv1file.cipher
    <libtakiyasha.qmc.qmcdataciphers.Mask128 object at 0x7ff682178c10>
    >>>

    - 打开一个外部 QMCv1 文件：
    >>> qmcv1file = QMCv1.open('/path/to/qmcv1file.qmcflac', mask=b'YourQMCv1Mask...')
    >>> qmcv1file
    <libtakiyasha.qmc.QMCv1 at 0x7ff6820d9e90, cipher <libtakiyasha.qmc.qmcdataciphers.Mask128 object at 0x7ff682178c10>, source '/path/to/qmcv1file.qmcflac'>
    >>>

    - 读取和写入，注意写入操作产生的修改需要调用 ``save()`` 方法显式保存：
    >>> qmcv1file.read(16)
    b'fLaC\\x00\\x00\\x00"\\x12\\x00\\x12\\x00\\x00\\x07)\\x00'
    >>> qmcv1file.seek(0, 2)
    36137109
    >>> qmcv1file.write(b'\\x00Writing something')
    18
    >>>

    - 保存上述操作产生的更改
    >>> # 如果该 QMCv1 对象不是从文件打开的，还需要 filething 参数
    >>> qmcv1file.save()
    >>>
    """

    @property
    def acceptable_ciphers(self):
        return [Mask128]

    @classmethod
    def from_file(cls,
                  qmcv1_filething: FilePath | IO[bytes], /,
                  master_key: BytesLike
                  ):
        """（已弃用，且将会在后续版本中删除。请尽快使用 ``QMCv1.open()`` 代替。）

        打开一个 QMCv1 文件或文件对象 ``qmcv1_filething``。

        第一个位置参数 ``qmcv1_filething`` 可以是文件路径（``str``、``bytes``
        或任何拥有方法 ``__fspath__()`` 的对象）。``qmcv1_filething``
        也可以是一个文件对象，但必须可读。

        第二个位置参数 ``master_key`` 用于解密音频数据，长度仅限 44、128 或 256 位。
        如果不符合长度要求，会触发 ``ValueError``。
        """
        warnings.warn(
            DeprecationWarning(
                f'{cls.__name__}.from_file() is deprecated, no longer used, '
                f'and may be removed in subsequent versions. '
                f'Use {cls.__name__}.open() instead.'
            )
        )
        return cls.open(qmcv1_filething, mask=master_key)

    @classmethod
    def open(cls,
             filething_or_info: tuple[Path | IO[bytes], QMCv1FileInfo | None] | FilePath | IO[bytes], /,
             mask: BytesLike
             ):
        """打开一个 QMCv1 文件，并返回一个 ``QMCv1`` 对象。

        第一个位置参数 ``filething`` 需要是一个文件路径或文件对象。
        可接受的文件路径类型包括：字符串、字节串、任何定义了 ``__fspath__()`` 方法的对象。
        如果是文件对象，那么必须可读且可寻址（其 ``seekable()`` 方法返回 ``True``）。

        第二个参数 ``mask`` 是必需的，用于主密钥。其长度必须为 44、128 或 256 位。

        Args:
            filething_or_info: 源文件的路径或文件对象
            mask: 文件的主密钥，其长度必须为 44、128 或 256 位
        Raises:
            ValueError: mask 的长度不符合上述要求
        """
        mask = tobytes(mask)

        def operation(fd: IO[bytes]) -> cls:
            if len(mask) == 44:
                cipher = Mask128.from_qmcv1_mask44(mask)
            elif len(mask) == 128:
                cipher = Mask128(mask)
            elif len(mask) == 256:
                cipher = Mask128.from_qmcv1_mask256(mask)
            else:
                raise ValueError(
                    f"the length of argument 'mask' must be 44, 128, or 256, not {len(mask)}"
                )

            fd.seek(fileinfo.cipher_data_offset, 0)
            return cls(cipher, fd.read(fileinfo.cipher_data_len))

        if isinstance(filething_or_info, tuple):
            filething_or_info: tuple[Path | IO[bytes], QMCv1FileInfo | None]
            if len(filething_or_info) != 2:
                raise TypeError(
                    "first argument 'filething_or_info' must be a file path, a file object, "
                    "or a tuple of probe_qmc(), probe_qmcv1() returns"
                )
            filething, fileinfo = filething_or_info
        else:
            filething, fileinfo = probe_qmcv1(filething_or_info)

        if fileinfo is None:
            raise CrypterCreatingError(
                f"{repr(filething)} is not a QMCv1 file"
            )
        elif not isinstance(fileinfo, QMCv1FileInfo):
            raise TypeError(
                f"second element of the tuple must be QMCv1FileInfo or None, not {type(fileinfo).__name__}"
            )

        if isfilepath(filething):
            with open(filething, mode='rb') as fileobj:
                instance = operation(fileobj)
                instance._name = Path(filething)
        else:
            fileobj = verify_fileobj(filething, 'binary',
                                     verify_readable=True,
                                     verify_seekable=True
                                     )
            fileobj_sourcefile = getattr(fileobj, 'name', None)
            instance = operation(fileobj)

            if fileobj_sourcefile is not None:
                instance._name = Path(fileobj_sourcefile)

        return instance

    def to_file(self, qmcv1_filething: FilePath | IO[bytes] = None) -> None:
        """（已弃用，且将会在后续版本中删除。请尽快使用 ``QMCv2.save()`` 代替。）

        将当前 QMCv1 对象的内容保存到文件 ``qmcv1_filething``。

        第一个位置参数 ``qmcv1_filething`` 可以是文件路径（``str``、``bytes``
        或任何拥有方法 ``__fspath__()`` 的对象）。``qmcv1_filething``
        也可以是一个文件对象，但必须可写。

        本方法会首先尝试写入 ``qmcv1_filething`` 指向的文件。
        如果未提供 ``qmcv1_filething``，则会尝试写入 ``self.name``
        指向的文件。如果两者都为空或未提供，则会触发 ``CrypterSavingError``。
        """
        warnings.warn(
            DeprecationWarning(
                f'{type(self).__name__}.from_file() is deprecated, no longer used, '
                f'and may be removed in subsequent versions. '
                f'Use {type(self).__name__}.save() instead.'
            )
        )
        return self.save(qmcv1_filething)

    def save(self, filething: FilePath | IO[bytes] = None) -> None:
        """将当前对象保存为一个新 QMCv1 文件。

        第一个参数 ``filething`` 是可选的，如果提供此参数，需要是一个文件路径或文件对象。
        可接受的文件路径类型包括：字符串、字节串、任何定义了 ``__fspath__()`` 方法的对象。
        如果是文件对象，那么必须可读且可寻址（其 ``seekable()`` 方法返回 ``True``）。

        Args:
            filething: 目标文件的路径或文件对象

        Raises:
            TypeError: 当前对象的属性 source 和参数 filething 都为空，无法保存文件
        """

        def operation(fd: IO[bytes]):
            fd.write(self.getvalue(nocryptlayer=True))

        if filething is None:
            if self.source is None:
                raise TypeError(
                    "attribute 'self.source' and argument 'filething' are empty, "
                    "don't know which file to save to"
                )
            filething = self.source

        if isfilepath(filething):
            with open(filething, mode='wb') as fileobj:
                return operation(fileobj)
        else:
            fileobj = verify_fileobj(filething, 'binary',
                                     verify_seekable=True,
                                     verify_writable=True
                                     )
            return operation(fileobj)

    @classmethod
    def new(cls, mask: BytesLike = None):
        """返回一个空 QMCv1 对象。

        第一个参数 ``mask`` 是可选的，如果提供，将被用作主密钥。
        """
        if mask is None:
            mask = make_salt(128)
        else:
            mask = tobytes(mask)
        return cls(Mask128(mask))


class QMCv2(EncryptedBytesIO):
    """基于 BytesIO 的 QMCv2 透明加密二进制流。

    所有读写相关方法都会经过透明加密层处理：
    读取时，返回解密后的数据；写入时，向缓冲区写入加密后的数据。

    调用读写相关方法时，附加参数 ``nocryptlayer=True``
    可绕过透明加密层，访问缓冲区内的原始加密数据。

    如果你要新建一个 QMCv2 对象，不要直接调用 ``__init__()``，而是使用构造器方法
    ``QMCv2.new()`` 和 ``QMCv2.open()`` 新建或打开已有 QMCv2 文件，
    使用已有 QMCv2 对象的 ``save()`` 方法将其保存到文件。

        使用示例：

    - 新建一个 QMCv2 对象以便编辑：
    >>> qmcv2file = QMCv2.new('mask')  # 必须选择使用的加密方式，可用值：'map'、'mask'、'rc4'
    >>> qmcv2file
    <libtakiyasha.qmc.QMCv2 at 0x7ff6820d9e90, cipher <libtakiyasha.qmc.qmcdataciphers.Mask128 object at 0x7ff682178c10>>
    >>>

    - 获取使用的主密钥：
    >>> qmcv2file.master_key  # 此处的密钥是随机生成的
    b'pdjx6epLTXuayfUKEWjxxcEWcWqybsMTTauxG7cIo5qsXH8...'
    >>>

    - 访问内部的 Cipher 对象：
    >>> qmcv2file.cipher
    <libtakiyasha.qmc.qmcdataciphers.Mask128 object at 0x7ff682178c10>
    >>>

    - 打开一个外部 QMCv2 文件：
    >>> qmcv2file = QMCv2.open('/path/to/qmcv2file-hardenedrc4.mflac0', core_key=b'YourQMCv2CoreKey')
    >>> qmcv2file
    <libtakiyasha.qmc.QMCv2 at 0x7ff672679e90, cipher <libtakiyasha.qmc.qmcdataciphers.HardenedRC4 object at 0x7ff6811ff6a0>, source '/path/to/qmcv2file-hardenedrc4.mflac0'>
    >>>
    根据文件使用的加密方式，你可能需要输入多个密钥。更多使用方法，请使用 ``help(QMCv2.open)`` 查看帮助。

    - 读取和写入，注意写入操作产生的修改需要调用 ``save()`` 方法显式保存：
    >>> qmcv2file.read(16)
    b'fLaC\\x00\\x00\\x00"\\x12\\x00\\x12\\x00\\x00\\x07)\\x00'
    >>> qmcv2file.seek(0, 2)
    36137109
    >>> qmcv2file.write(b'\\x00Writing something')
    18
    >>>

    - 保存上述操作产生的更改
    >>> # 如果该 QMCv2 对象不是从文件打开的，还需要 filething 参数
    >>> qmcv2file.save(core_key=b'YourQMCv2CoreKey')
    >>>
    根据你想要的输出文件加密方式，你可能需要输入多个密钥。更多使用方法，请使用 ``help(QMCv2.save)`` 查看帮助。

    - 访问源文件末尾的附加数据（QTag 或 STag），如果没有，则为 ``None``：
    >>> qmcv2file.extra_info
    QMCv2QTag(song_id=0, unknown=2)
    >>> qmcv2file.extra_info.song_id
    1145141919810
    >>> qmcv2file.extra_info.unknown
    2
    >>>
    """

    @property
    def acceptable_ciphers(self):
        return [HardenedRC4, Mask128]

    def __init__(self, cipher: HardenedRC4 | Mask128, /, initial_bytes: BytesLike = b'') -> None:
        """基于 BytesIO 的 QMCv2 透明加密二进制流。

        所有读写相关方法都会经过透明加密层处理：
        读取时，返回解密后的数据；写入时，向缓冲区写入加密后的数据。

        调用读写相关方法时，附加参数 ``nocryptlayer=True``
        可绕过透明加密层，访问缓冲区内的原始加密数据。

        如果你要新建一个 QMCv2 对象，不要直接调用 ``__init__()``，而是使用构造器方法
        ``QMCv2.new()`` 和 ``QMCv2.open()`` 新建或打开已有 QMCv2 文件，
        使用已有 QMCv2 对象的 ``save()`` 方法将其保存到文件。

        Args:
            cipher: 要使用的 cipher，必须是一个 libtakiyasha.qmc.qmcdataciphers.Mask128/HardenedRC4 对象
            initial_bytes: 内置缓冲区的初始数据
        """
        super().__init__(cipher, initial_bytes)

        self._extra_info: QMCv2QTag | QMCv2STag | None = None

        self._core_key_deprecated: bytes | None = None
        self._garble_key1_deprecated: bytes | None = None
        self._garble_key2_deprecated: bytes | None = None

    @property
    def extra_info(self) -> QMCv2QTag | QMCv2STag | None:
        """源文件末尾的附加信息（如果有），根据类型可分为 QTag 或 STag。"""
        return self._extra_info

    @extra_info.setter
    def extra_info(self, value: QMCv2QTag | QMCv2STag) -> None:
        """源文件末尾的附加信息（如果有），根据类型可分为 QTag 或 STag。"""
        if isinstance(value, (QMCv2QTag, QMCv2STag)):
            self._extra_info = value
        elif value is None:
            raise TypeError(
                f"None cannot be assigned to attribute 'extra_info'. "
                f"Use `del self.extra_info` instead"
            )
        else:
            raise TypeError(
                f"attribute 'extra_info' must be QMCv2QTag or QMCv2STag, not {repr(value)}"
            )

    @extra_info.deleter
    def extra_info(self) -> None:
        """源文件末尾的附加信息（如果有），根据类型可分为 QTag 或 STag。"""
        self._extra_info = None

    @property
    def master_key(self) -> bytes | None:
        if isinstance(self.cipher, Mask128):
            ret = self.cipher.getkey('original')
            if ret:
                return ret

        return super().master_key

    @property
    def core_key(self) -> bytes | None:
        """（已弃用，且将会在后续版本中删除。）

        核心密钥，用于加/解密主密钥。

        ``QMCv2.from_file()`` 会在当前对象被创建时设置此属性；而 ``QMCv2.open()`` 则不会。
        """
        warnings.warn(
            DeprecationWarning(
                f'{type(self).__name__}.core_key or {type(self).__name__}.simple_key'
                f'is deprecated, no longer used, '
                f'and may be removed in subsequent versions. '
                f'You need to manage the core key by your self.'
            )
        )
        return self._core_key_deprecated

    @core_key.setter
    def core_key(self, value: BytesLike) -> None:
        """（已弃用，且将会在后续版本中删除。）

        核心密钥，用于加/解密主密钥。

        ``QMCv2.from_file()`` 会在当前对象被创建时设置此属性；而 ``QMCv2.open()`` 则不会。
        """
        warnings.warn(
            DeprecationWarning(
                f'{type(self).__name__}.core_key or {type(self).__name__}.simple_key'
                f'is deprecated, no longer used, '
                f'and may be removed in subsequent versions. '
                f'You need to manage the core key by your self.'
            )
        )
        if value is None:
            raise TypeError(
                f"None cannot be assigned to attribute 'core_key'. "
                f"Use `del self.core_key` instead"
            )
        self._core_key_deprecated = tobytes(value)

    @core_key.deleter
    def core_key(self) -> None:
        """（已弃用，且将会在后续版本中删除。）

        核心密钥，用于加/解密主密钥。

        ``QMCv2.from_file()`` 会在当前对象被创建时设置此属性；而 ``QMCv2.open()`` 则不会。
        """
        warnings.warn(
            DeprecationWarning(
                f'{type(self).__name__}.core_key or {type(self).__name__}.simple_key'
                f'is deprecated, no longer used, '
                f'and may be removed in subsequent versions. '
                f'You need to manage the core key by your self.'
            )
        )
        self._core_key_deprecated = None

    simple_key = core_key

    @property
    def garble_key1(self) -> bytes | None:
        """（已弃用，且将会在后续版本中删除。）

        混淆密钥 1，用于加/解密主密钥。

        ``QMCv2.from_file()`` 会在当前对象被创建时设置此属性；而 ``QMCv2.open()`` 则不会。
        """
        warnings.warn(
            DeprecationWarning(
                f'{type(self).__name__}.garble_key1 or {type(self).__name__}.mix_key1'
                f'is deprecated, no longer used, '
                f'and may be removed in subsequent versions. '
                f'You need to manage garble keys by your self.'
            )
        )
        return self._garble_key1_deprecated

    @garble_key1.setter
    def garble_key1(self, value: BytesLike) -> None:
        """（已弃用，且将会在后续版本中删除。）

        混淆密钥 1，用于加/解密主密钥。

        ``QMCv2.from_file()`` 会在当前对象被创建时设置此属性；而 ``QMCv2.open()`` 则不会。
        """
        warnings.warn(
            DeprecationWarning(
                f'{type(self).__name__}.garble_key1 or {type(self).__name__}.mix_key1'
                f'is deprecated, no longer used, '
                f'and may be removed in subsequent versions. '
                f'You need to manage garble keys by your self.'
            )
        )
        if value is None:
            raise TypeError(
                f"None cannot be assigned to attribute 'garble_key1'. "
                f"Use `del self.garble_key1` instead"
            )
        self._garble_key1_deprecated = tobytes(value)

    @garble_key1.deleter
    def garble_key1(self):
        """（已弃用，且将会在后续版本中删除。）

        混淆密钥 1，用于加/解密主密钥。

        ``QMCv2.from_file()`` 会在当前对象被创建时设置此属性；而 ``QMCv2.open()`` 则不会。
        """
        warnings.warn(
            DeprecationWarning(
                f'{type(self).__name__}.garble_key1 or {type(self).__name__}.mix_key1'
                f'is deprecated, no longer used, '
                f'and may be removed in subsequent versions. '
                f'You need to manage garble keys by your self.'
            )
        )
        self._garble_key1_deprecated = None

    mix_key1 = garble_key1

    @property
    def garble_key2(self) -> bytes | None:
        """（已弃用，且将会在后续版本中删除。）

        混淆密钥 2，用于加/解密主密钥。

        ``QMCv2.from_file()`` 会在当前对象被创建时设置此属性；而 ``QMCv2.open()`` 则不会。
        """
        warnings.warn(
            DeprecationWarning(
                f'{type(self).__name__}.garble_key2 or {type(self).__name__}.mix_key2'
                f'is deprecated, no longer used, '
                f'and may be removed in subsequent versions. '
                f'You need to manage garble keys by your self.'
            )
        )
        return self._garble_key2_deprecated

    @garble_key2.setter
    def garble_key2(self, value: BytesLike) -> None:
        """（已弃用，且将会在后续版本中删除。）

        混淆密钥 2，用于加/解密主密钥。

        ``QMCv2.from_file()`` 会在当前对象被创建时设置此属性；而 ``QMCv2.open()`` 则不会。
        """
        warnings.warn(
            DeprecationWarning(
                f'{type(self).__name__}.garble_key2 or {type(self).__name__}.mix_key2'
                f'is deprecated, no longer used, '
                f'and may be removed in subsequent versions. '
                f'You need to manage garble keys by your self.'
            )
        )
        if value is None:
            raise TypeError(
                f"None cannot be assigned to attribute 'garble_key2'. "
                f"Use `del self.garble_key2` instead"
            )
        self._garble_key2_deprecated = tobytes(value)

    @garble_key2.deleter
    def garble_key2(self):
        """（已弃用，且将会在后续版本中删除。）

        混淆密钥 2，用于加/解密主密钥。

        ``QMCv2.from_file()`` 会在当前对象被创建时设置此属性；而 ``QMCv2.open()`` 则不会。
        """
        warnings.warn(
            DeprecationWarning(
                f'{type(self).__name__}.garble_key2 or {type(self).__name__}.mix_key2'
                f'is deprecated, no longer used, '
                f'and may be removed in subsequent versions. '
                f'You need to manage garble keys by your self.'
            )
        )
        self._garble_key2_deprecated = None

    mix_key2 = garble_key2

    @classmethod
    def from_file(cls,
                  qmcv2_filething: FilePath | IO[bytes], /,
                  simple_key: BytesLike = None,
                  mix_key1: BytesLike = None,
                  mix_key2: BytesLike = None, *,
                  master_key: BytesLike = None,
                  ):
        """（已弃用，且将会在后续版本中删除。请尽快使用 ``QMCv2.open()`` 代替。）

        打开一个 QMCv2 文件或文件对象 ``qmcv2_filething``。

        第一个位置参数 ``qmcv2_filething`` 可以是文件路径（``str``、``bytes``
        或任何拥有方法 ``__fspath__()`` 的对象）。``qmcv2_filething``
        也可以是一个文件对象，但必须可读、可跳转（``qmcv2_filething.seekable() == True``）。

        本方法会寻找文件内嵌主密钥的位置和加密方式，进而判断所用加密算法的类型。

        如果提供了参数 ``master_key``，那么此参数将会被视为主密钥，
        用于判断加密算法类型和解密音频数据，同时会跳过其他步骤。
        其必须是类字节对象，且转换为 ``bytes`` 的长度必须是 128、256
        或 512 位。如果不符合长度要求，会触发 ``ValueError``。否则：

        - 如果未能找到文件内嵌的主密钥，那么参数 ``master_key`` 是必需的。
        - 如果文件内嵌的主密钥，其加密版本为 V1，那么参数 ``simple_key`` 是必需的。
        - 如果文件内嵌的主密钥，其加密版本为 V2，那么除了 ``simple_key``，参数``mix_key1``、``mix_key2`` 也是必需的。

        以上特定条件中的必需参数，如果缺失，则会触发 ``ValueError``。
        """
        warnings.warn(
            DeprecationWarning(
                f'{cls.__name__}.from_file() is deprecated, no longer used, '
                f'and may be removed in subsequent versions. '
                f'Use {cls.__name__}.open() instead.'
            )
        )
        instance = cls.open(qmcv2_filething,
                            core_key=simple_key,
                            garble_keys=(mix_key1, mix_key2),
                            master_key=master_key
                            )
        instance._core_key_deprecated = tobytes(simple_key)
        instance._garble_key1_deprecated = tobytes(mix_key1)
        instance._garble_key2_deprecated = tobytes(mix_key2)

    @classmethod
    def open(cls,
             filething_or_info: tuple[Path | IO[bytes], QMCv2FileInfo | None] | FilePath | IO[bytes], /,
             core_key: BytesLike = None,
             garble_keys: Iterable[BytesLike] = None,
             master_key: BytesLike = None,
             encrypt_method: Literal['map', 'mask', 'rc4'] = None
             ):
        """打开一个 QMCv2 文件，并返回一个 ``QMCv2`` 对象。

        第一个位置参数 ``filething_or_info`` 需要是一个文件路径或文件对象。
        可接受的文件路径类型包括：字符串、字节串、任何定义了 ``__fspath__()`` 方法的对象。
        如果是文件对象，那么必须可读且可寻址（其 ``seekable()`` 方法返回 ``True``）。

        ``filething_or_info`` 也可以接受 ``probe_qmc()`` 和 ``probe_qmcv2()`` 函数的返回值：
        一个包含两个元素的元组，第一个元素是源文件的路径或文件对象，第二个元素是源文件的信息。

        第二个参数 ``core_key`` 一般情况下是必需的，用于解密文件内嵌的主密钥。
        例外：如果你提供了第五个参数 ``master_key``，那么它是可选的。

        第三个参数 ``garble_keys`` 是一组混淆密钥，仅在探测到文件内嵌的主密钥使用了
        V2 加密时是必需的。如果提供，应当是一个由至少一个混淆密钥（字节对象）组成的可迭代对象，
        且其中的混淆密钥需要按照正确的顺序排列，否则会导致无法解密主密钥和打开文件。
        在其他情况下，此参数会被忽略。

        第四个参数 ``master_key`` 可选，如果提供，将会被作为主密钥使用，
        而文件内置的主密钥会被忽略，``core_key``、``garble_keys`` 也不再是必需参数。
        例外：如果探测到文件未嵌入任何形式的密钥，那么此参数是必需的。

        第五个参数 ``encrypt_method`` 用于指定文件数据使用的加密方式，支持以下值：

        - ``'map'`` 或 ``'mask'`` - 掩码表（Mask128）
        - ``'rc4'`` - 强化版 RC4（HardenedRC4）
        - ``None`` - 不指定，由 ``probe_qmcv2()`` 自行探测

        此参数的设置会覆盖 ``probe_qmc()`` 或 ``probe_qmcv2()`` 的探测结果。

        Args:
            filething_or_info: 源文件的路径或文件对象，或者 probe_qmc() 和 probe_qmcv2() 的返回值
            core_key: 核心密钥，用于解密文件内嵌的主密钥
            garble_keys: 混淆密钥列表，用于解密使用 V2 加密的主密钥，其中的混淆密钥需要按照正确的顺序排列
            master_key: 如果提供，将会被作为主密钥使用，而文件内置的主密钥会被忽略
            encrypt_method: 用于指定文件数据使用的加密方式，支持 'map'、'mask'、'rc4' 或 None
        Raises:
            TypeError: 参数 core_key 和 master_key 都未提供，或者缺少 garble_keys 从而无法解密 V2 加密的主密钥
            ValueError: encrypt_method 的值不符合上述条件
            CrypterCreatingError: probe_qmcv2() 返回的文件信息中，master_key_encryption_ver 的值是当前不支持的
        """
        if core_key is not None:
            core_key = tobytes(core_key)
        if master_key is not None:
            master_key = tobytes(master_key)
        if encrypt_method is not None:
            if encrypt_method not in ('map', 'mask', 'rc4'):
                if isinstance(encrypt_method, str):
                    raise ValueError(
                        f"argument 'encrypt_method' must be 'map', 'mask', or 'rc4', "
                        f"not {repr(encrypt_method)}"
                    )
                else:
                    raise TypeError(
                        f"argument 'encrypt_method' must be str, "
                        f"not {type(encrypt_method).__name__}"
                    )
        if garble_keys is not None:
            garble_keys = [tobytes(_) for _ in garble_keys]

        def operation(fd: IO[bytes]) -> cls:
            cipher_data_len = fileinfo.cipher_data_len
            extra_info = fileinfo.extra_info
            master_key_encrypted = fileinfo.master_key_encrypted
            master_key_encryption_ver = fileinfo.master_key_encryption_ver
            cipher_ctor = fileinfo.cipher_ctor

            if master_key is None:
                if isinstance(extra_info, QMCv2STag):
                    raise TypeError(
                        "argument 'master_key' is required to "
                        "QMCv2 file ends with STag"
                    )
                if core_key is None:
                    raise TypeError(
                        "argument 'core_key' is required to "
                        "decrypt the protected master key"
                    )
                if master_key_encryption_ver == 1:
                    target_master_key = QMCv2KeyEncryptV1(core_key).decrypt(
                        master_key_encrypted
                    )
                elif master_key_encryption_ver == 2:
                    if not garble_keys:
                        raise ValueError(
                            "argument 'garble_keys' is required to decrypt the "
                            "QMCv2 Key Encryption V2 protected master key"
                        )
                    target_master_key = QMCv2KeyEncryptV2(core_key, garble_keys).decrypt(master_key_encrypted)
                else:
                    raise CrypterCreatingError(
                        f"unsupported master key encryption version {master_key_encryption_ver}"
                    )
            else:
                target_master_key = master_key
                cipher_ctor = _guess_cipher_ctor(target_master_key, is_encrypted=False)

            if encrypt_method in ('map', 'mask'):
                cipher_ctor = Mask128
            elif encrypt_method == 'rc4':
                cipher_ctor = HardenedRC4

            if cipher_ctor is None:
                raise TypeError(
                    "don't know which cipher to use, "
                    f"please try {cls.__name__}.open() again "
                    f"with argument 'encrypt_method'"
                )

            cipher = cipher_ctor(target_master_key)
            fd.seek(0, 0)
            inst = cls(cipher, fd.read(cipher_data_len))
            inst._extra_info = extra_info

            return inst

        if isinstance(filething_or_info, tuple):
            filething_or_info: tuple[Path | IO[bytes], QMCv2FileInfo | None]
            if len(filething_or_info) != 2:
                raise TypeError(
                    "first argument 'filething_or_info' must be a file path, a file object, "
                    "or a tuple of probe_qmc(), probe_qmcv2() returns"
                )
            filething, fileinfo = filething_or_info
        else:
            filething, fileinfo = probe_qmcv2(filething_or_info)

        if fileinfo is None:
            raise CrypterCreatingError(
                f"{repr(filething)} is not a QMCv2 file"
            )
        elif not isinstance(fileinfo, QMCv2FileInfo):
            raise TypeError(
                f"second element of the tuple must be QMCv2FileInfo or None, not {type(fileinfo).__name__}"
            )

        if isfilepath(filething):
            with open(filething, mode='rb') as fileobj:
                instance = operation(fileobj)
                instance._name = Path(filething)
        else:
            fileobj = verify_fileobj(filething, 'binary',
                                     verify_readable=True,
                                     verify_seekable=True
                                     )
            fileobj_sourcefile = getattr(fileobj, 'name', None)
            instance = operation(fileobj)

            if fileobj_sourcefile is not None:
                instance._name = Path(fileobj_sourcefile)

        return instance

    def to_file(self,
                qmcv2_filething: FilePath | IO[bytes] = None, /,
                tag_type: Literal['qtag', 'stag'] = None,
                simple_key: BytesLike = None,
                master_key_enc_ver: IntegerLike = 1,
                mix_key1: BytesLike = None,
                mix_key2: BytesLike = None
                ) -> None:
        """（已弃用，且将会在后续版本中删除。请尽快使用 ``QMCv2.save()`` 代替。）

        将当前 QMCv2 对象的内容保存到文件 ``qmcv2_filething``。

        第一个位置参数 ``qmcv2_filething`` 可以是文件路径（``str``、``bytes``
        或任何拥有方法 ``__fspath__()`` 的对象）。``qmcv2_filething``
        也可以是一个文件对象，但必须可写。

        本方法会首先尝试写入 ``qmcv2_filething`` 指向的文件。
        如果未提供 ``qmcv2_filething``，则会尝试写入 ``self.name``
        指向的文件。如果两者都为空或未提供，则会触发 ``CrypterSavingError``。

        参数 ``tag_type`` 决定在文件末尾附加的内容，仅支持以下值：
            - ``None`` - 将主密钥加密后直接附加在文件末尾。
            - ``qtag`` - 将主密钥加密后封装在 QTag 信息中，附加在文件末尾。
            - ``stag`` - 将 STag 信息附加在文件末尾。
                - 注意：选择 STag 意味着文件内不会内嵌主密钥，你需要自己记下主密钥。
                - 访问属性 ``self.master_key`` 获取主密钥。

        如果 ``tag_type`` 为其他值，会触发 ``ValueError``。

        无论 ``tag_type`` 为何值（``stag`` 除外），都需要使用 ``simple_key`` 加密主密钥。
        如果参数 ``master_key_enc_ver=2``，还需要 ``mix_key1`` 和 ``mix_key2``。
        如果未提供这些参数，则会使用当前 QMCv2 对象的同名属性代替。
        如果两者都为 ``None`` 或未提供，则会触发 ``CrypterSavingError``。
        """
        warnings.warn(
            DeprecationWarning(
                f'{type(self).__name__}.from_file() is deprecated, no longer used, '
                f'and may be removed in subsequent versions. '
                f'Use {type(self).__name__}.save() instead.'
            )
        )
        with_extra_info = False
        if isinstance(self.extra_info, (QMCv2QTag, QMCv2QTag)) and tag_type:
            with_extra_info = True
        if master_key_enc_ver == 1:
            mix_key1 = None
            mix_key2 = None
        elif master_key_enc_ver == 2:
            if mix_key1 is None:
                mix_key1 = self.garble_key1
            if mix_key2 is None:
                mix_key2 = self.garble_key2
            if mix_key1 is None and mix_key2 is None:
                raise TypeError(
                    "argument 'mix_key1' and 'mix_key2' is required to "
                    "decrypt the QMCv2 Key Encryption V2 protected master key"
                )
            elif mix_key1 is None:
                raise TypeError(
                    "argument 'mix_key1' is required to "
                    "decrypt the QMCv2 Key Encryption V2 protected master key"
                )
            elif mix_key2 is None:
                raise TypeError(
                    "argument 'mix_key2' is required to "
                    "decrypt the QMCv2 Key Encryption V2 protected master key"
                )
        else:
            raise ValueError("argument 'master_key_enc_ver' must be 1 or 2, "
                             f"not {master_key_enc_ver}"
                             )
        if simple_key is None:
            simple_key = self.core_key
        return self.save(core_key=simple_key,
                         filething=qmcv2_filething,
                         garble_keys=(mix_key1, mix_key2),
                         with_extra_info=with_extra_info
                         )

    def save(self,
             core_key: BytesLike = None,
             filething: FilePath | IO[bytes] = None,
             garble_keys: Iterable[BytesLike] = None,
             with_extra_info: bool = False
             ) -> None:
        """将当前对象保存为一个新 QMCv2 文件。

        第一个参数 ``core_key`` 一般是必需的，用于加密主密钥，以便嵌入到文件。
        例外：参数 ``with_extra_info=True`` 且当前对象的属性 ``extra_info``
        是一个 ``QMCv2STag`` 对象，此时它是可选的，其值会被忽略。

        第二个参数 ``filething`` 是可选的，如果提供此参数，需要是一个文件路径或文件对象。
        可接受的文件路径类型包括：字符串、字节串、任何定义了 ``__fspath__()`` 方法的对象。
        如果是文件对象，那么必须可读且可寻址（其 ``seekable()`` 方法返回 ``True``）。
        如果未提供此参数，那么将会尝试使用当前对象的 ``source`` 属性；如果后者也不可用，则引发
        ``TypeError``。

        第三个参数 ``garble_keys`` 是一组混淆密钥，如果提供，将会对主密钥使用 V2 加密，
        否则使用 V1 加密。它应当是一个由至少一个混淆密钥（字节对象）组成的可迭代对象，
        且其中的混淆密钥需要按照正确的顺序排列。
        如果参数 ``with_extra_info=True`` 且当前对象的属性 ``extra_info``
        是一个 ``QMCv2STag`` 对象，此参数会被忽略。

        第四个参数 ``with_extra_info`` 如果为 ``True``，且当前对象的属性 ``extra_info`` 是
        ``QMCv2QTag`` 或 ``QMCv2STag`` 对象，那么这些对象将会被序列化后嵌入文件。

        Args:
            core_key: 核心密钥，用于加密主密钥，以便嵌入到文件
            filething: 目标文件的路径或文件对象
            garble_keys: 混淆密钥列表，用于使用 V2 方式加密主密钥，其中的混淆密钥需要按照正确的顺序排列
            with_extra_info: 是否在文件末尾添加额外信息（self.extra_info）

        Raises:
            TypeError: 当前对象的属性 source 和参数 filething 都为空，无法保存文件；参数 core_key 和 master_key 都未提供，或者缺少 garble_key1 或 garble_key2 用于使用 V2 方式加密主密钥
        """
        if core_key is not None:
            core_key = tobytes(core_key)
        if garble_keys is not None:
            garble_keys = [tobytes(_) for _ in garble_keys]

        def operation(fd: IO[bytes]) -> None:
            fd.seek(0, 0)
            extra_info = self.extra_info

            if with_extra_info:
                if isinstance(extra_info, QMCv2STag):
                    warnings.warn(
                        CrypterSavingWarning(
                            "Extra info (self.extra_info) will be export to STag data, "
                            "which cannot save the master key. "
                            "So you should save the master key in other way. "
                            "Use 'self.master_key' to get it."
                        )
                    )
                    tag_serialized = extra_info.dump()
                    fd.write(self.getvalue(nocryptlayer=True))
                    fd.write(tag_serialized)
                    fd.write(len(tag_serialized).to_bytes(4, 'big'))
                    fd.write(b'STag')

                    return

            master_key = self.master_key
            if core_key is None:
                raise TypeError(
                    "argument 'core_key' is required to encrypt the master key "
                    "before embed to file"
                )
            if with_extra_info:
                if isinstance(extra_info, QMCv2QTag):
                    master_key_encrypted = QMCv2KeyEncryptV1(core_key).encrypt(master_key)
                    master_key_encrypted_b64encoded = b64encode(master_key_encrypted)
                    tag_serialized = extra_info.dump(master_key_encrypted_b64encoded)
                    fd.write(self.getvalue(nocryptlayer=True))
                    fd.write(tag_serialized)
                    fd.write(len(tag_serialized).to_bytes(4, 'big'))
                    fd.write(b'QTag')

                    return

            if garble_keys:
                master_key_encrypted = QMCv2KeyEncryptV2(core_key, garble_keys).encrypt(master_key)
                master_key_encrypted_b64encoded = b64encode(
                    b'QQMusic EncV2,Key:' + master_key_encrypted
                )
            else:
                master_key_encrypted = QMCv2KeyEncryptV1(core_key).encrypt(master_key)
                master_key_encrypted_b64encoded = b64encode(master_key_encrypted)
            fd.write(self.getvalue(nocryptlayer=True))
            fd.write(master_key_encrypted_b64encoded)
            fd.write(len(master_key_encrypted_b64encoded).to_bytes(4, 'little'))

            return

        if filething is None:
            if self.source is None:
                raise TypeError(
                    "attribute 'self.source' and argument 'filething' are empty, "
                    "don't know which file to save to"
                )
            filething = self.source

        if isfilepath(filething):
            with open(filething, mode='wb') as fileobj:
                return operation(fileobj)
        else:
            fileobj = verify_fileobj(filething, 'binary',
                                     verify_seekable=True,
                                     verify_writable=True
                                     )
            return operation(fileobj)

    @classmethod
    def new(cls, encrypt_method: Literal['map', 'mask', 'rc4'], /):
        """返回一个空 QMCv2 对象。

        第一个位置参数 ``encrypt_method`` 是必需的，用于指示使用的加密方式，支持以下值：

        - ``'map'`` 或 ``'mask'`` - 掩码表（Mask128）
        - ``'rc4'`` - 强化版 RC4（HardenedRC4）

        Raises:
            ValueError: encrypt_method 的值不符合上述条件
        """
        if encrypt_method in ('map', 'mask'):
            cipher = Mask128.from_qmcv2_key256(make_random_ascii_string(256).encode('ascii'))
        elif encrypt_method == 'rc4':
            cipher = HardenedRC4(make_random_ascii_string(512).encode('ascii'))
        elif isinstance(encrypt_method, str):
            raise ValueError(
                f"argument 'encrypt_method' must be 'map', 'mask', or 'rc4', "
                f"not {repr(encrypt_method)}"
            )
        else:
            raise TypeError(
                f"argument 'encrypt_method' must be str, "
                f"not {type(encrypt_method).__name__}"
            )

        return cls(cipher)



# -*- coding: utf-8 -*-
from __future__ import annotations

from functools import wraps
from pathlib import Path
from typing import Callable, IO, Iterable, Mapping, NamedTuple, TypeVar

from .typedefs import BytesLike, FilePath, KT, T, VT
from .typeutils import isfilepath, tobytes, verify_fileobj

__all__ = [
    'BINARIES_ROOTDIR',
    'bytestrxor',
    'getattribute',
    'proberfuncfactory'
]

BINARIES_ROOTDIR = Path(__file__).parent / 'binaries'


def getattribute(obj: object,
                 name: str,
                 *default: T,
                 follow_callable: bool = False,
                 callable_args: Iterable = None,
                 callable_kwargs: Mapping | Iterable[tuple[KT, VT]] = None
                 ) -> T:
    """用法：``getattribute(object, name[, default][, follow_callable=...][, callable_args=...][, callable_kwargs=...]) -> value``

    ``getattribute()`` 的用法与 ``getattr()`` 基本一致。

    返回对象 ``obj`` 具名属性 ``name`` 的值。``name`` 必须是字符串。
    如果该字符串是对象的属性之一，则返回该属性的值。

    例如，``getattribute(x, 'foobar')`` 等同于 ``x.foobar``。

    如果指定的属性不存在，且提供了 ``default`` 值，则返回它，否则触发 ``AttributeError``。

    如果向 ``getattribute()`` 提供了关键字参数 ``follow_callable=True``，而且
    ``x.foobar`` 存在并且是一个可调对象（例如目标对象的一个方法），那么 ``getattribute()``
    会尝试获取该属性的返回值，作为 ``getattribute()`` 的返回值。

    Args:
        obj: 目标对象
        name: 目标属性的名字
        default: 可选，在目标对象不存在目标属性时，返回的默认值
        follow_callable: 若为真值，则在目标属性为可调对象时，返回调用该属性后的返回值
        callable_args: 可选，若目标属性为可调对象且
            follow_callable 为真值，调用该属性时传入的位置参数
        callable_kwargs: 可选，若目标属性为可调对象且
            follow_callable 为真值，调用该属性时传入的关键字参数
    """
    if len(default) > 1:
        raise TypeError(f'getattribute expected at most 3 arguments, got {2 + len(default)}')
    else:
        attr = getattr(obj, name, *default)

    if callable_args is None:
        callable_args = ()
    else:
        callable_args = tuple(callable_kwargs)
    if callable_kwargs is None:
        callable_kwargs = {}
    else:
        callable_kwargs = dict(callable_kwargs)

    if callable(attr) and bool(follow_callable):
        return attr(*callable_args, **callable_kwargs)
    return attr


def bytestrxor(term1: BytesLike, term2: BytesLike, /) -> bytes:
    """用法：``bytestrxor(term1, term2) -> xored_bytes``

    返回两个字节对象或类字节对象 ``term1`` 和 ``term2`` 经过异或之后的结果。

    ``term1`` 和 ``term2`` 在转换为 ``bytes`` 之后的长度必须相等，否则会触发
    ``ValueError``。
    """
    bytestring1 = tobytes(term1)
    bytestring2 = tobytes(term2)

    if len(bytestring1) != len(bytestring2):
        raise ValueError(
            'only byte strings of equal length can be xored: '
            f'term 1 ({len(bytestring1)}) != term 2 ({len(bytestring2)})'
        )

    return bytes(b1 ^ b2 for b1, b2 in zip(bytestring1, bytestring2))


_FileInfo = TypeVar('_FileInfo', bound=NamedTuple)


def proberfuncfactory(docstring_from=None):
    """本装饰器是一个工厂函数，被装饰的函数必须具备以下行为：

    - 至少接受一个参数（位于第一个），并将此参数视为文件对象进行操作
    - 内含探测文件结构的所有逻辑
    - 如果探测到受支持的文件，返回一个 NamedTuple 对象，该对象包含探测到的信息；否则返回 `None`

    具备以上行为的函数在被本装饰器装饰后，将会变成一个文件结构探测函数，具备以下行为：

    - 第一个纯位置参数为文件路径或文件对象，随后的任何参数都会被传递给被装饰的函数
    - 返回一个 2 元组：

        - 第一个元素是接受的文件路径或文件对象
        - 第二个元素是探测结果，具体内容见上文
    """

    def prober(
            operation: Callable[[IO[bytes], ...], _FileInfo | None]
    ) -> Callable[[FilePath, ...], tuple[FilePath | IO[bytes], _FileInfo | None]]:
        @wraps(operation)
        def wrapper(filething: FilePath | IO[bytes], /, *args, **kwargs) -> tuple[FilePath | IO[bytes], _FileInfo | None]:
            if isfilepath(filething):
                with open(filething, mode='rb') as fileobj:
                    return Path(filething), operation(fileobj, *args, **kwargs)
            else:
                fileobj = verify_fileobj(filething, 'binary',
                                         verify_readable=True,
                                         verify_seekable=True
                                         )
                fileobj_origpos = fileobj.tell()
                prs = operation(fileobj, *args, **kwargs)
                fileobj.seek(fileobj_origpos, 0)

                return fileobj, prs

        if isinstance(docstring_from, str):
            wrapper.__doc__ = str(docstring_from)
        elif docstring_from is not None:
            docstring = getattr(docstring_from, '__doc__', None)
            if docstring is not None:
                wrapper.__doc__ = docstring

        return wrapper

    if callable(docstring_from):
        return prober(docstring_from)
    return prober



# -*- coding: utf-8 -*-
from __future__ import annotations


class LibTakiyashaException(Exception):
    pass


class CipherEncryptingError(LibTakiyashaException):
    pass


class CipherDecryptingError(LibTakiyashaException):
    pass


class CrypterCreatingError(LibTakiyashaException):
    pass


class CrypterSavingError(LibTakiyashaException):
    pass



# -*- coding: utf-8 -*-
from __future__ import annotations

from typing import Generator, Literal

from ..miscutils import bytestrxor
from ..prototypes import KeyStreamBasedStreamCipherSkel
from ..typedefs import BytesLike, IntegerLike
from ..typeutils import tobytes, toint

__all__ = ['Mask32', 'Mask32FromRecipe']


class Mask32(KeyStreamBasedStreamCipherSkel):
    def __init__(self, mask32: BytesLike, /) -> None:
        self._mask32 = tobytes(mask32)
        if len(self._mask32) != 32:
            raise ValueError(f"invalid mask length: should be 32, got {len(self._mask32)}")

    def getkey(self, keyname: str = 'master') -> bytes | None:
        if keyname == 'master':
            return self._mask32

    @classmethod
    def cls_keystream(cls, mask32: BytesLike, nbytes: IntegerLike, offset: IntegerLike, /) -> Generator[int, None, None]:
        offset = toint(offset)
        nbytes = toint(nbytes)
        if offset < 0:
            raise ValueError("third argument 'offset' must be a non-negative integer")
        if nbytes < 0:
            raise ValueError("second argument 'nbytes' must be a non-negative integer")
        maskblk_data: bytes = tobytes(mask32)
        maskblk_len = len(maskblk_data)
        if maskblk_len != 32:
            raise ValueError(f"invalid mask length: should be 32, not {maskblk_len}")

        target_in_maskblk_len = nbytes
        target_offset_in_maskblk = offset % maskblk_len
        if target_offset_in_maskblk == 0:
            target_before_maskblk_area_len = 0
        else:
            target_before_maskblk_area_len = maskblk_len - target_offset_in_maskblk
        yield from maskblk_data[target_offset_in_maskblk:target_offset_in_maskblk + target_before_maskblk_area_len]
        target_in_maskblk_len -= target_before_maskblk_area_len

        target_overrided_whole_maskblk_count = target_in_maskblk_len // maskblk_len
        target_after_maskblk_area_len = target_in_maskblk_len % maskblk_len

        for _ in range(target_overrided_whole_maskblk_count):
            yield from maskblk_data
        yield from maskblk_data[:target_after_maskblk_area_len]

    def keystream(self,
                  operation: Literal['encrypt', 'decrypt'],
                  nbytes: IntegerLike,
                  offset: IntegerLike, /
                  ) -> Generator[int, None, None]:
        yield from self.cls_keystream(self._mask32, nbytes, offset)


class Mask32FromRecipe(Mask32):
    def __init__(self, recipe: BytesLike, core_key: BytesLike, /) -> None:
        recipe = tobytes(recipe)
        core_key = tobytes(core_key)

        for varname, var, expectlen in ('core_key', core_key, 32), ('recipe', recipe, 8):
            if len(var) != expectlen:
                f"invalid length of argument '{varname}': should be {expectlen}, not {len(var)}"

        mask_recipe_unpacked = int.from_bytes(recipe, 'little')
        mask_stage1 = str(mask_recipe_unpacked).encode('ascii')
        if len(mask_stage1) >= 32:
            mask_stage2 = mask_stage1[:32]
        else:
            mask_stage2_pad_len = 32 - len(mask_stage1)
            mask_stage2_stage1_fullpad_count = (mask_stage2_pad_len // len(mask_stage1))
            mask_stage2_stage1_fullpad_len = len(mask_stage1) * mask_stage2_stage1_fullpad_count
            mask_stage2_remain_len = mask_stage2_pad_len - mask_stage2_stage1_fullpad_len

            mask_stage2_composition = [mask_stage1]
            for _ in range(mask_stage2_stage1_fullpad_count):
                mask_stage2_composition.append(mask_stage1)
            mask_stage2_composition.append(mask_stage1[:mask_stage2_remain_len])
            mask_stage2 = b''.join(mask_stage2_composition)

        mask_final = bytestrxor(mask_stage2, core_key)

        self._source_recipe = recipe
        self._core_key = core_key

        super().__init__(mask_final)

    def getkey(self, keyname: str = 'master') -> bytes | None:
        if keyname == 'original':
            return self._source_recipe
        elif keyname == 'core':
            return self._core_key
        else:
            return super().getkey(keyname)



# -*- coding: utf-8 -*-
from __future__ import annotations

from hashlib import md5
from typing import Generator, Literal

from ..prototypes import KeyStreamBasedStreamCipherSkel
from ..typedefs import BytesLike, IntegerLike
from ..typeutils import CachedClassInstanceProperty, tobytes, toint

__all__ = ['KGMCryptoLegacy']


def kugou_md5sum(data: bytes, /) -> bytes:
    md5sum = md5(data)
    md5digest = md5sum.digest()
    ret = bytearray(md5sum.digest_size)
    for i in range(0, md5sum.digest_size, 2):
        ret[i] = md5digest[14 - i]
        ret[i + 1] = md5digest[14 + 1 - i]

    return bytes(ret)


class KGMCryptoLegacy(KeyStreamBasedStreamCipherSkel):
    @CachedClassInstanceProperty
    def keysize(self) -> int:
        return 17

    @CachedClassInstanceProperty
    def tablesize(self) -> int:
        return 17 * 16

    def __init__(self,
                 table1: BytesLike,
                 table2: BytesLike,
                 tablev2: BytesLike,
                 core_key_test_data: BytesLike, /,
                 vpr_key: BytesLike = None,
                 ) -> None:
        self._table1 = tobytes(table1)
        self._table2 = tobytes(table2)
        self._tablev2 = tobytes(tablev2)

        for idx, valname_val in enumerate([('table1', self._table1),
                                           ('table2', self._table2),
                                           ('tablev2', self._tablev2)],
                                          start=1
                                          ):
            valname, val = valname_val
            if len(val) != self.tablesize:
                raise ValueError(
                    f"invalid length of position {idx} argument '{valname}': "
                    f"should be {self.tablesize}, not {len(val)}"
                )

        self._core_key_test_data = tobytes(core_key_test_data)
        if len(self._core_key_test_data) != self.keysize:
            raise ValueError(
                f"invalid length of fourth argument 'core_key_test_data': "
                f"should be {self.keysize}, not {len(self._core_key_test_data)}"
            )
        if vpr_key is None:
            self._vpr_key = None
        else:
            self._vpr_key = tobytes(vpr_key)
            if len(self._vpr_key) != self.keysize:
                raise ValueError(
                    f"invalid length of argument 'vpr_key': "
                    f"should be {self.keysize}, not {len(self._vpr_key)}"
                )

    def getkey(self, keyname: str = 'master') -> bytes | None:
        if keyname == 'master':
            return self._core_key_test_data
        elif keyname == 'table1':
            return self._table1
        elif keyname == 'table2':
            return self._table2
        elif keyname == 'tablev2':
            return self._tablev2
        elif keyname == 'vprkey':
            return self._vpr_key

    def prexor_encrypt(self, data: BytesLike, offset: IntegerLike, /) -> Generator[int, None, None]:
        offset = toint(offset)
        vpr_key = self._vpr_key
        keysize = self.keysize
        for idx, byte in enumerate(data, start=offset):
            if vpr_key is not None:
                byte ^= vpr_key[idx % keysize]
            yield byte ^ ((byte % 16) << 4)

    @staticmethod
    def prexor_decrypt(data: BytesLike, offset: IntegerLike, /) -> Generator[int, None, None]:
        offset = toint(offset)
        for idx, byte in enumerate(data, start=offset):
            yield byte ^ ((byte % 16) << 4)

    def postxor_decrypt(self, data: BytesLike, offset: IntegerLike, /) -> Generator[int, None, None]:
        offset = toint(offset)
        vpr_key = self._vpr_key
        keysize = self.keysize
        for idx, byte in enumerate(data, start=offset):
            if vpr_key is None:
                yield byte
            else:
                yield byte ^ vpr_key[idx % keysize]

    def genmask(self,
                nbytes: IntegerLike,
                offset: IntegerLike, /
                ) -> Generator[int, None, None]:
        nbytes = toint(nbytes)
        offset = toint(offset)
        if offset < 0:
            raise ValueError("second argument 'offset' must be a non-negative integer")
        if nbytes < 0:
            raise ValueError("first argument 'nbytes' must be a non-negative integer")

        tablesize: int = self.tablesize
        table1 = self._table1
        table2 = self._table2
        tablev2 = self._tablev2
        for idx in range(offset, offset + nbytes):
            idx_urs4 = idx >> 4
            value = 0
            while idx_urs4 >= 17:
                value ^= table1[idx_urs4 % tablesize]
                idx_urs4 >>= 4
                value ^= table2[idx_urs4 % tablesize]
                idx_urs4 >>= 4
            yield value ^ tablev2[idx % tablesize]

    def keystream(self,
                  operation: Literal['encrypt', 'decrypt'],
                  nbytes: IntegerLike,
                  offset: IntegerLike, /
                  ) -> Generator[int, None, None]:
        ck_test_data = self._core_key_test_data
        keysize: int = self.keysize

        mask_strm = self.genmask(nbytes, offset)
        if operation == 'encrypt':
            for idx, msb in enumerate(mask_strm, start=offset):
                yield msb ^ ck_test_data[idx % keysize]
        elif operation == 'decrypt':
            for idx, msb in enumerate(mask_strm, start=offset):
                msb ^= ck_test_data[idx % keysize]
                yield msb ^ ((msb % 16) << 4)
        elif isinstance(operation, str):
            raise ValueError(
                f"first argument 'operation' must be 'encrypt' or 'decrypt', not {operation}"
            )
        else:
            raise TypeError(
                f"first argument 'operation' must be str, not {type(operation).__name__}"
            )



# -*- coding: utf-8 -*-
from __future__ import annotations

import json
import warnings
from base64 import b64decode, b64encode
from dataclasses import asdict, dataclass, field as dcfield
from pathlib import Path
from secrets import token_bytes
from typing import Callable, IO, Literal, NamedTuple, Type, overload

from mutagen import flac, id3

from .exceptions import CrypterCreatingError
from .keyutils import make_random_ascii_string, make_random_number_string
from .miscutils import BINARIES_ROOTDIR, bytestrxor, proberfuncfactory
from .prototypes import EncryptedBytesIO
from .stdciphers import ARC4, StreamedAESWithModeECB
from .typedefs import BytesLike, FilePath
from .typeutils import isfilepath, tobytes, verify_fileobj
from .warns import CrypterCreatingWarning

warnings.filterwarnings(action='default', category=DeprecationWarning, module=__name__)

__all__ = ['CloudMusicIdentifier', 'NCM', 'probe_ncm', 'NCMFileInfo']

MODULE_BINARIES_ROOTDIR = BINARIES_ROOTDIR / Path(__file__).stem


@dataclass(init=True)
class CloudMusicIdentifier:
    """解析、储存和重建网易云音乐 163key 。

    如果 163key 的来源文件来自网易云音乐客户端，根据你使用的客户端平台和版本，
    一些字段可能会缺失，从而被设为默认值。

    可以按照操作数据类（``dataclass``）实例的方式操作本类的实例。
    """

    def __post_init__(self) -> None:
        self._orig_ncm_tag: dict | None = None
        self._orig_ncm_163key: bytes | None = None
        self._orig_tag_key: bytes | None = None

    format: str = ''
    """文件格式，通常为 ``mp3`` 或 ``flac``。"""
    musicId: str = ''
    """此歌曲在网易云音乐的 ID。"""
    musicName: str = ''
    """此歌曲的标题。"""
    artist: list[list[str | int]] = dcfield(default_factory=list)
    """此歌曲的歌手。"""
    album: str = ''
    """此歌曲所属的专辑。"""
    albumId: int = 0
    """此歌曲所属的专辑在网易云音乐平台的 ID。"""
    albumPicDocId: int = 0
    """此歌曲所属的专辑，其封面在网易云音乐平台的 ID。"""
    albumPic: str = ''
    """此歌曲所属的专辑，其封面的下载链接。"""
    mvId: int = 0
    """此歌曲的 MV 在网易云音乐平台的 ID。"""
    flag: int = 0
    bitrate: int = 0
    """此歌曲的比特率。"""
    duration: int = 0
    """此歌曲的长度（单位为秒）。"""
    gain: float = 0.0
    """此歌曲的平均响度。"""
    mp3DocId: str = ''
    alias: list[str] = dcfield(default_factory=list)
    """此歌曲在网易云音乐平台的别名。"""
    transNames: list[str] = dcfield(default_factory=list)
    """此歌曲标题在网易云音乐平台的翻译。"""

    def to_mutagen_tag(self,
                       tag_type: Literal['FLAC', 'ID3'] = None,
                       with_ncm_163key: bool = True,
                       tag_key: BytesLike | None = None,
                       return_cached_first: bytes = True
                       ) -> flac.FLAC | id3.ID3:
        """将 CloudMusicIdentifier 对象导出为 Mutagen 库使用的标签格式实例：
        ``mutagen.flac.FLAC`` 和 ``mutagen.id3.ID3``。

        ``tag_type`` 用于选择需要导出为何种格式的标签实例，仅支持 ``FLAC``
        和 ``ID3``。如果留空，则根据 ``self.format`` 决定。如果两者都为空，则会触发 ``ValueError``。

        Args:
            tag_type: 需要导出为何种格式的标签实例，仅支持 'FLAC' 和 'ID3'
            with_ncm_163key: 是否在导出的标签中嵌入 163key
            tag_key: （仅当 with_163key=True）歌曲信息密钥，用于加密 163key，以便将其写入注释
            return_cached_first: （仅当 with_163key=True）在满足特定条件时，将缓存的 163key 写入注释，而不是重新生成一个

        Examples:
            >>> from mutagen import flac, mp3
            [...]
            >>> ncm_tag1: CloudMusicIdentifier
            >>> ncm_tag1.format
            'flac'
            >>> mutagen_tag1 = ncm_tag1.to_mutagen_tag()
            >>> type(mutagen_tag1)
            mutagen.flac.FLAC
            >>> flactag = flac.FLAC('test.flac')
            >>> flactag.update(mutagen_tag1)
            >>> flactag.save()
            >>>
            [...]
            >>> ncm_tag2: CloudMusicIdentifier
            >>> ncm_tag2.format
            'mp3'
            >>> mutagen_tag2 = ncm_tag2.to_mutagen_tag()
            >>> type(mutagen_tag2)
            mutagen.id3.ID3
            >>> mp3tag = mp3.MP3('test.mp3')
            >>> mp3tag.update(mutagen_tag2)
            >>> mp3tag.save()
            >>>
        """
        if tag_type is None:
            if self.format.lower() == 'flac':
                tag_type = 'FLAC'
            elif self.format.lower() == 'mp3':
                tag_type = 'ID3'
            elif not self.format:
                raise ValueError(
                    "don't know which type of tag is needed: "
                    "self.format and 'tag_type' are empty"
                )
            else:
                raise ValueError(
                    "don't know which type of tag is needed: "
                    "'tag_type' is empty, and the value of self.format is not supported"
                )
        if tag_type == 'FLAC':
            with open(MODULE_BINARIES_ROOTDIR / 'empty.flac', mode='rb') as _f:
                # 受 mutagen 功能限制，编辑 FLAC 标签之前必须打开一个空 FLAC 文件
                tag: flac.FLAC | id3.ID3 = flac.FLAC(_f)
            keymaps = {
                'musicName': ('title', lambda _: [_]),
                'artist'   : ('artist', lambda _: list(str(list(__)[0]) for __ in list(_))),
                'album'    : ('album', lambda _: [_])
            }
        elif tag_type == 'ID3':
            tag: flac.FLAC | id3.ID3 = id3.ID3()
            keymaps = {
                'musicName': ('TIT2', lambda _: id3.TIT2(text=[_], encoding=3)),
                'artist'   : ('TPE1', lambda _: id3.TPE1(text=list(str(list(__)[0]) for __ in list(_)), encoding=3)),
                'album'    : ('TALB', lambda _: id3.TALB(text=[_], encoding=3))
            }
        else:
            raise ValueError(
                f"'tag_type' must be 'FLAC', 'ID3', or None, not {repr(tag_type)}"
            )

        tagkey_constructor: tuple[str, Callable[[str], list[str]] | Callable[[str], id3.Frame]]
        for attrname, tagkey_constructor in keymaps.items():
            tagkey, constructor = tagkey_constructor
            attr = getattr(self, attrname)
            if attr:
                tag[tagkey] = constructor(attr)

        if with_ncm_163key:
            ncm_163key = self.to_ncm_163key(tag_key=tag_key,
                                            return_cached_first=return_cached_first
                                            )
            if isinstance(tag, flac.FLAC):
                tag['description'] = [ncm_163key.decode('ascii')]
            elif isinstance(tag, id3.ID3):
                tag['TXXX::comment'] = id3.TXXX(encoding=3, desc='comment', text=[ncm_163key.decode('ascii')])

        return tag

    @classmethod
    def from_ncm_163key(cls, ncm_163key: str | BytesLike, /, tag_key: BytesLike = None):
        """将一个 163key 字符串/字节对象转换为 CloudMusicIdentifier 对象。

        本方法会缓存给定的 163key，以及该 163key 解密后的结果，用于确保 ``self.to_ncm_163key()``
        返回值的一致性。

        Args:
            ncm_163key: 以“163key”开头的字符串/字节对象
            tag_key: 歌曲信息密钥，用于解密 163key
        """
        if isinstance(ncm_163key, str):
            ncm_163key = bytes(ncm_163key, encoding='utf-8')
        else:
            ncm_163key = tobytes(ncm_163key)
        if tag_key is None:
            tag_key = b'\x23\x31\x34\x6c\x6a\x6b\x5f\x21\x5c\x5d\x26\x30\x55\x3c\x27\x28'
        else:
            tag_key = tobytes(tag_key)

        ncm_tag_serialized_encrypted_encoded = ncm_163key[22:]  # 去除开头的 b"163 key(Don't modify):"
        ncm_tag_serialized_encrypted = b64decode(ncm_tag_serialized_encrypted_encoded, validate=True)
        ncm_tag_serialized = StreamedAESWithModeECB(tag_key).decrypt(
            ncm_tag_serialized_encrypted
        )[6:]  # 去除字节串开头的 b'music:'
        ncm_tag = json.loads(ncm_tag_serialized)

        instance = cls(**ncm_tag)
        instance._orig_ncm_tag = ncm_tag
        instance._orig_ncm_163key = ncm_163key
        instance._orig_tag_key = tag_key
        return instance

    def to_ncm_163key(self,
                      tag_key: BytesLike = None,
                      return_cached_first: bytes = True
                      ) -> bytes:
        """将 CloudMusicIdentifier 对象导出为 163key。

        第一个参数 ``tag_key`` 用于解密 163key。如果留空，则使用默认值：
        ``b'\x23\x31\x34\x6c\x6a\x6b\x5f\x21\x5c\x5d\x26\x30\x55\x3c\x27\x28'``

        第二个参数 ``return_cached`` 如果为 ``True``，
        那么在当前对象转换而来的字典（下称当前字典）与缓存的 163key 解密得到的字典（下称缓存字典）
        满足以下条件时，本方法会直接返回 ``self.from_ncm_163key()`` 缓存的 163key：

        - 当前字典包含了缓存字典中的所有字段，且在两个字典中，这些键对应的值也是一致的
        - 当前字典中缓存字典没有的字段，其值为默认值（空值）

        如果以上条件中的任意一条未被满足，转而返回一个根据当前对象重新生成的 163key。
        Args:
            tag_key: 歌曲信息密钥，用于加密 163key
            return_cached_first: 在满足特定条件时，返回缓存的 163key，而不是重新生成一个
        """
        if tag_key is None:
            tag_key = b'\x23\x31\x34\x6c\x6a\x6b\x5f\x21\x5c\x5d\x26\x30\x55\x3c\x27\x28'
        else:
            tag_key = tobytes(tag_key)

        if return_cached_first and self._orig_ncm_tag:
            target_ncm_tag = {_ck: _cv for _ck, _cv in asdict(self).items() if _cv or _ck in self._orig_ncm_tag}
        else:
            target_ncm_tag = asdict(self)

        def operation() -> bytes:
            ncm_tag_serialized = json.dumps(target_ncm_tag, ensure_ascii=False).encode('utf-8')
            ncm_tag_serialized_encrypted = StreamedAESWithModeECB(tag_key).encrypt(b'music:' + ncm_tag_serialized)
            ncm_163key = b"163 key(Don't modify):" + b64encode(ncm_tag_serialized_encrypted)

            return ncm_163key

        if return_cached_first and tag_key == self._orig_tag_key and self._orig_ncm_tag and self._orig_ncm_163key:
            if len(target_ncm_tag) == len(self._orig_ncm_tag):
                for k, v in self._orig_ncm_tag.items():
                    if target_ncm_tag[k] != v:
                        break
                else:
                    return tobytes(self._orig_ncm_163key)

        return operation()

    def to_mutagen_style_dict(self):
        """（已弃用，且将会在后续版本中删除。请尽快使用
        ``CloudMusicIdentifier.to_mutagen_tag()`` 代替，以便极大简化步骤。）

        根据当前对象储存的解析结果，构建并返回一个 Mutagen VorbisComment/ID3 风格的字典。

        此方法需要当前对象的 ``format`` 属性来决定构建何种风格的字典，
        并且只支持 ``'flac'`` (VorbisComment) 和 ``'mp3'`` (ID3)。

        本方法不支持嵌入封面图像，你需要通过其他手段做到。

        配合 Mutagen 使用（以 FLAC 为例）：

        >>> from mutagen import flac  # type: ignore
        >>> ncm_tag = CloudMusicIdentifier(format='flac')
        >>> mutagen_flac = mutagen.flac.FLAC('target.flac')  # type: ignore
        >>> mutagen_flac.clear()  # 可选，此步骤将会清空 mutagen_flac 已有的数据
        >>> mutagen_flac.update(ncm_tag.to_mutagen_style_dict())
        >>> mutagen_flac.save()
        >>>

        配合 Mutagen 使用（以 MP3 为例，稍微麻烦一些）：

        >>> from mutagen import id3, mp3  # type: ignore
        >>> ncm_tag = CloudMusicIdentifier(format='mp3')
        >>> mutagen_mp3 = mutagen.mp3.MP3('target.mp3')  # type: ignore
        >>> mutagen_mp3.clear()  # 可选，此步骤将会清空 mutagen_mp3 已有的数据
        >>> for key, value in ncm_tag.to_mutagen_style_dict().items():
        ...     id3frame_cls = getattr(id3, key[:4])
        ...     id3frame = mutagen_mp3.get(key)
        ...     if id3frame is None:
        ...         mutagen_mp3[key] = id3frame_cls(text=value, desc='comment')
        ...     elif id3frame.text:
        ...         id3frame.text = value
        ...         mutagen_mp3[key] = id3frame
        ...
        >>> mutagen_mp3.save()
        >>>
        """
        warnings.warn(
            DeprecationWarning(
                f'{type(self).__name__}.to_mutagen_style_dict() is deprecated, no longer used, '
                f'and may be removed in subsequent versions. '
                f'Use {type(self).__name__}.to_mutagen_tag() instead.'
            )
        )

        comment = self.to_ncm_163key().decode('utf-8')
        if not isinstance(self.format, str):
            raise TypeError(f"'self.format' must be str, not {type(self.format)}")
        elif self.format.lower() == 'flac':
            ret = {
                'title' : [self.musicName],
                'artist': [artistinfo[0] for artistinfo in self.artist],
                'album' : [self.album],
            }
            if comment is not None:
                ret['comment'] = [comment]
        elif self.format.lower() == 'mp3':
            ret = {
                'TIT2': [self.musicName],
                'TPE1': [artistinfo[0] for artistinfo in self.artist],
                'TALB': [self.album]
            }
            if comment is not None:
                ret['TXXX::comment'] = [comment]
        else:
            raise ValueError(f"unsupported tag format '{self.format}'")

        return ret


class NCMFileInfo(NamedTuple):
    """用于储存 NCM 文件的信息。

    要想打开文件，需要使用 ``NCM.open()``。不可以直接通过 ``NCMFileInfo`` 打开文件。
    """
    master_key_encrypted: bytes
    """受加密保护的主密钥。"""
    ncm_163key: bytes
    """歌曲在网易云音乐平台上的歌曲信息，为 AES 加密的 JSON 字符串。"""
    cipher_ctor: Callable[[...], ARC4]
    """Cipher 的构造函数，接受一个必需参数（密钥），返回一个用于解密数据的
    Cipher 对象。通常就是 Cipher 类自身。"""
    cipher_data_offset: int
    """加密数据在文件中开始的位置。"""
    cipher_data_len: int
    """加密数据在文件中的长度。"""
    cover_data_offset: int
    """封面数据在文件中的位置。"""
    cover_data_len: int
    """封面数据在文件中的长度。"""
    opener: Callable[[tuple[FilePath | IO[bytes], NCMFileInfo] | FilePath | IO[bytes], ...], NCM]
    """打开文件的方式，为一个可调对象，其会返回一个加密文件对象。"""
    opener_kwargs_required: tuple[str, ...]
    """通过 ``opener`` 打开文件时，所必需的关键字参数的名称。"""
    opener_kwargs_optional: tuple[str, ...]
    """通过 ``opener`` 打开文件时，可选的关键字参数的名称。
    
    此属性仅储存可能会影响 ``opener`` 行为的可选关键字参数；
    对 ``opener`` 行为没有影响的可选关键字参数不会出现在此属性中。
    """


@overload
def probe_ncm(filething: FilePath | IO[bytes], /) -> tuple[FilePath | IO[bytes], NCMFileInfo | None]:
    pass


@proberfuncfactory
def probe_ncm(filething, /):
    """用法：probe_ncm(filething)

    探测源文件 ``filething`` 是否为一个 NCM 文件。

    返回一个 2 个元素长度的元组：

    - 第一个元素为 ``filething``
    - 如果 ``filething`` 是 NCM 文件，那么第二个元素为一个 ``NCMFileInfo`` 对象；
    - 否则为 ``None``。

    本方法的返回值可以用于 ``NCM.open()`` 的第一个位置参数。

    Args:
        filething: 源文件的路径或文件对象
    Returns:
        一个 2 个元素长度的元组：第一个元素为 filething；如果
        filething 是 NCM 文件，那么第二个元素为一个 NCMFileInfo 对象；否则为 None。
    """
    filething.seek(0, 0)

    if not filething.read(10).startswith(b'CTENFDAM'):
        return

    master_key_encrypted_xored_len = int.from_bytes(filething.read(4), 'little')
    master_key_encrypted_xored = filething.read(master_key_encrypted_xored_len)
    master_key_encrypted = bytestrxor(b'd' * master_key_encrypted_xored_len,
                                      master_key_encrypted_xored
                                      )

    ncm_163key_xored_len = int.from_bytes(filething.read(4), 'little')
    ncm_163key_xored = filething.read(ncm_163key_xored_len)
    ncm_163key = bytestrxor(b'c' * ncm_163key_xored_len, ncm_163key_xored)

    filething.seek(5, 1)

    cover_space_len = int.from_bytes(filething.read(4), 'little')
    cover_data_len = int.from_bytes(filething.read(4), 'little')
    if cover_space_len - cover_data_len < 0:
        raise CrypterCreatingError(f'file structure error: '
                                   f'cover space length ({cover_space_len}) '
                                   f'< cover data length ({cover_data_len})'
                                   )
    cover_data_offset = filething.tell()
    cipher_data_offset = filething.seek(cover_space_len, 1)
    cipher_data_len = filething.seek(0, 2) - cipher_data_offset

    return NCMFileInfo(
        master_key_encrypted=master_key_encrypted,
        ncm_163key=ncm_163key,
        cipher_ctor=ARC4,
        cipher_data_offset=cipher_data_offset,
        cipher_data_len=cipher_data_len,
        cover_data_offset=cover_data_offset,
        cover_data_len=cover_data_len,
        opener=NCM.open,
        opener_kwargs_required=('core_key',),
        opener_kwargs_optional=('tag_key', 'master_key')
    )


class NCM(EncryptedBytesIO):
    """基于 BytesIO 的 NCM 透明加密二进制流。

    所有读写相关方法都会经过透明加密层处理：
    读取时，返回解密后的数据；写入时，向缓冲区写入加密后的数据。

    调用读写相关方法时，附加参数 ``nocryptlayer=True``
    可绕过透明加密层，访问缓冲区内的原始加密数据。

    如果你要新建一个 NCM 对象，不要直接调用 ``__init__()``，而是使用构造器方法
    ``NCM.new()`` 和 ``NCM.open()`` 新建或打开已有 NCM 文件，
    使用已有 NCM 对象的 ``save()`` 方法将其保存到文件。

    使用示例：

    - 新建一个 NCM 对象以便编辑：
    >>> ncmfile = NCM.new()
    >>> ncmfile
    <libtakiyasha.ncm.NCM at 0x7f26c4f4a390, cipher <libtakiyasha.stdciphers.ARC4 object at 0x7f26c51214b0>>
    >>>

    - 获取使用的主密钥：
    >>> ncmfile.master_key  # 此处的密钥是随机生成的
    b'60564957557881842441053098814sAwKPjVq2gK9JXW0nV7BF3iQXh0J1ra34dh9UqfiMCUUzOrcyKERif9IfdFf5toJk6rO8TZaVSYkQVtZClVY'
    >>>

    - 访问内部的 Cipher 对象：
    >>> ncmfile.cipher
    <libtakiyasha.stdciphers.ARC4 object at 0x7f26c51214b0>
    >>>

    - 打开一个外部 NCM 文件：
    >>> ncmfile = NCM.open('/path/to/ncmfile.ncm', core_key=b'YourNCMCoreKey', tag_key=b'YourNCMTagKey')
    >>> ncmfile
    <libtakiyasha.ncm.NCM at 0x7f26c44e5080, cipher <libtakiyasha.stdciphers.ARC4 object at 0x7f26c4ef1270>, source '/path/to/ncmfile.ncm'>
    >>>
    更多使用方法，请使用 ``help(NCM.open)`` 查看帮助。

    - 读取和写入，注意写入操作产生的修改需要调用 ``save()`` 方法显式保存：
    >>> ncmfile.read(16)
    b'fLaC\\x00\\x00\\x00"\\x12\\x00\\x12\\x00\\x00\\x07)\\x00'
    >>> ncmfile.seek(0, 2)
    36137109
    >>> ncmfile.write(b'\\x00Writing something')
    18
    >>>

    - 保存上述操作产生的更改
    >>> # 如果该 NCM 对象不是从文件打开的，还需要 filething 参数
    >>> ncmfile.save(core_key=b'YourNCMCoreKey', tag_key=b'YourNCMTagKey')
    >>>
    更多使用方法，请使用 ``help(NCM.save)`` 查看帮助。

    - 获取歌曲标签和封面信息，详见 ``CloudMusicIdentifier``：
    >>> ncmfile.ncm_tag
    CloudMusicIdentifier(format=..., musicId=..., musicName=..., artist=[[..., ...], [..., ...]], album=..., albumId=..., albumPicDocId=..., albumPic=..., mvId=..., flag=..., bitrate=..., duration=..., gain=..., mp3DocId=..., alias=[...], transNames=[...])
    >>> len(ncmfile.cover_data)
    141248
    >>>
    """

    @classmethod
    def from_file(cls,
                  ncm_filething: FilePath | IO[bytes], /,
                  core_key: BytesLike,
                  ):
        """（已弃用，且将会在后续版本中删除。请尽快使用 ``NCM.open()`` 代替。）

        打开一个已有的 NCM 文件 ``ncm_filething``。

        第一个位置参数 ``ncm_filething`` 可以是 ``str``、``bytes`` 或任何拥有 ``__fspath__``
        属性的路径对象。``ncm_filething`` 也可以是文件对象，该对象必须可读和可跳转
        （``ncm_filething.seekable() == True``）。

        本方法需要在文件中寻找并解密主密钥，随后使用主密钥解密音频数据。

        核心密钥 ``core_key`` 是第二个参数，用于解密找到的主密钥。
        """
        warnings.warn(
            DeprecationWarning(
                f'{cls.__name__}.from_file() is deprecated, no longer used, '
                f'and may be removed in subsequent versions. '
                f'Use {cls.__name__}.open() instead.'
            )
        )
        return cls.open(ncm_filething, core_key=core_key)

    @classmethod
    def open(cls,
             filething_or_info: tuple[Path | IO[bytes], NCMFileInfo | None] | FilePath | IO[bytes], /,
             core_key: BytesLike = None,
             tag_key: BytesLike = None,
             master_key: BytesLike = None
             ):
        """打开一个 NCM 文件，并返回一个 ``NCM`` 对象。

        第一个位置参数 ``filething_or_info`` 需要是一个文件路径或文件对象。
        可接受的文件路径类型包括：字符串、字节串、任何定义了 ``__fspath__()`` 方法的对象。
        如果是文件对象，那么必须可读且可寻址（其 ``seekable()`` 方法返回 ``True``）。

        ``filething_or_info`` 也可以接受 ``probe_ncm()`` 函数的返回值：
        一个包含两个元素的元组，第一个元素是源文件的路径或文件对象，第二个元素是源文件的信息。

        第二个参数 ``core_key`` 一般情况下是必需的，用于解密文件内嵌的主密钥。
        例外：如果你提供了第四个参数 ``master_key``，那么它是可选的。

        第三个参数 ``tag_key`` 可选，用于解密文件内嵌的歌曲信息。如果留空，则使用默认值：
        ``b'\x23\x31\x34\x6c\x6a\x6b\x5f\x21\x5c\x5d\x26\x30\x55\x3c\x27\x28'``

        第四个参数 ``master_key`` 可选，如果提供，将会被作为主密钥使用，
        而文件内置的主密钥会被忽略，``core_key`` 也不再是必需参数。
        一般不需要填写此参数，因为 NCM 文件总是内嵌加密的主密钥，从而可以轻松地获得。

        Args:
            filething_or_info: 源文件的路径或文件对象，或者 probe_ncm() 的返回值
            core_key: 核心密钥，用于解密文件内嵌的主密钥
            tag_key: 歌曲信息密钥，用于解密文件内嵌的歌曲信息
            master_key: 如果提供，将会被作为主密钥使用，而文件内置的主密钥会被忽略
        Raises:
            TypeError: 参数 core_key 和 master_key 都未提供
        """
        if core_key is not None:
            core_key = tobytes(core_key)
        if tag_key is not None:
            tag_key = tobytes(tag_key)
        if master_key is not None:
            master_key = tobytes(master_key)
        if master_key is None and core_key is None:
            raise TypeError(
                f"{cls.__name__}.open() missing 1 argument: 'core_key'"
            )

        def operation(fd: IO[bytes]) -> cls:
            if master_key is None:
                target_master_key = StreamedAESWithModeECB(core_key).decrypt(
                    fileinfo.master_key_encrypted
                )[17:]  # 去除开头的 b'neteasecloudmusic'
            else:
                target_master_key = master_key
            cipher = fileinfo.cipher_ctor(target_master_key)

            try:
                ncm_tag = CloudMusicIdentifier.from_ncm_163key(
                    fileinfo.ncm_163key,
                    tag_key=tag_key
                )
            except Exception as exc:
                warnings.warn(f'skip parsing 163key, because an exception was raised while parsing: '
                              f'{type(exc).__name__}: {exc}',
                              CrypterCreatingWarning
                              )
                warnings.warn(f"you may need to check if the file {repr(filething)} "
                              f"is corrupted.",
                              CrypterCreatingWarning
                              )
                ncm_tag = CloudMusicIdentifier()

            fd.seek(fileinfo.cover_data_offset, 0)
            cover_data = fd.read(fileinfo.cover_data_len)

            fd.seek(fileinfo.cipher_data_offset, 0)
            initial_bytes = fd.read(fileinfo.cipher_data_len)

            inst = cls(cipher, initial_bytes)
            inst._cover_data = cover_data
            inst._ncm_tag = ncm_tag

            return inst

        if isinstance(filething_or_info, tuple):
            filething_or_info: tuple[Path | IO[bytes], NCMFileInfo | None]
            if len(filething_or_info) != 2:
                raise TypeError(
                    "first argument 'filething_or_info' must be a file path, a file object, "
                    "or a tuple of probe_ncm() returns"
                )
            filething, fileinfo = filething_or_info
        else:
            filething, fileinfo = probe_ncm(filething_or_info)

        if fileinfo is None:
            raise CrypterCreatingError(
                f"{repr(filething)} is not a NCM file"
            )
        elif not isinstance(fileinfo, NCMFileInfo):
            raise TypeError(
                f"second element of the tuple must be NCMFileInfo or None, not {type(fileinfo).__name__}"
            )

        if isfilepath(filething):
            with open(filething, mode='rb') as fileobj:
                instance = operation(fileobj)
                instance._name = Path(filething)
        else:
            fileobj = verify_fileobj(filething, 'binary',
                                     verify_readable=True,
                                     verify_seekable=True
                                     )
            fileobj_sourcefile = getattr(fileobj, 'name', None)
            instance = operation(fileobj)

            if fileobj_sourcefile is not None:
                instance._name = Path(fileobj_sourcefile)

        return instance

    def to_file(self,
                ncm_filething: FilePath | IO[bytes] = None, /,
                core_key: BytesLike = None
                ) -> None:
        """（已弃用，且将会在后续版本中删除。请尽快使用 ``NCM.save()`` 代替。）

        将当前 NCM 对象保存到文件 ``filething``。
        此过程会向 ``ncm_filething`` 写入 NCM 文件结构。

        第一个位置参数 ``ncm_filething`` 可以是 ``str``、``bytes`` 或任何拥有 ``__fspath__``
        属性的路径对象。``ncm_filething`` 也可以是文件对象，该对象必须可写。

        第二个位置参数 ``core_key`` 是可选的。
        如果提供此参数，本方法会将其作为核心密钥来加密主密钥；否则，使用
        ``self.core_key`` 代替。如果两者都为 ``None`` 或未提供，触发 ``ValueError``。

        如果提供了 ``ncm_filething``，本方法将会把数据写入 ``ncm_filething``
        指向的文件。否则，本方法以写入模式打开一个指向 ``self.name``
        的文件对象，将数据写入此文件对象。如果两者都为空或未提供，则会触发
        ``ValueError``。
        """
        warnings.warn(
            DeprecationWarning(
                f'{type(self).__name__}.from_file() is deprecated, no longer used, '
                f'and may be removed in subsequent versions. '
                f'Use {type(self).__name__}.save() instead.'
            )
        )
        if not core_key:
            core_key = self.core_key
        return self.save(core_key, filething=ncm_filething)

    def save(self,
             core_key: BytesLike,
             filething: FilePath | IO[bytes] = None,
             tag_key: BytesLike | None = None
             ) -> None:
        """将当前对象保存为一个新 NCM 文件。

        第一个参数 ``core_key`` 是必需的，用于加密主密钥，以便嵌入到文件。

        第二个参数 ``filething`` 是可选的，如果提供此参数，需要是一个文件路径或文件对象。
        可接受的文件路径类型包括：字符串、字节串、任何定义了 ``__fspath__()`` 方法的对象。
        如果是文件对象，那么必须可读且可寻址（其 ``seekable()`` 方法返回 ``True``）。
        如果未提供此参数，那么将会尝试使用当前对象的 ``source`` 属性；如果后者也不可用，则引发
        ``TypeError``。

        第三个参数 ``tag_key`` 可选，用于加密歌曲信息，以便嵌入到文件。如果留空，则使用默认值：
        ``b'\x23\x31\x34\x6c\x6a\x6b\x5f\x21\x5c\x5d\x26\x30\x55\x3c\x27\x28'``

        Args:
            core_key: 核心密钥，用于加密主密钥，以便嵌入到文件
            filething: 目标文件的路径或文件对象
            tag_key: 歌曲信息密钥，用于加密歌曲信息，以便嵌入到文件
        """
        core_key = tobytes(core_key)
        if tag_key is not None:
            tag_key = tobytes(tag_key)

        def operation(fd: IO[bytes]) -> None:
            fd.seek(0, 0)

            fd.write(b'CTENFDAM')
            fd.seek(2, 1)

            master_key = self.master_key
            master_key_encrypted = StreamedAESWithModeECB(core_key).encrypt(b'neteasecloudmusic' + master_key)
            master_key_encrypted_xored = bytestrxor(b'd' * len(master_key_encrypted), master_key_encrypted)
            master_key_encrypted_xored_len = len(master_key_encrypted_xored)
            fd.write(master_key_encrypted_xored_len.to_bytes(4, 'little'))
            fd.write(master_key_encrypted_xored)

            ncm_163key = self.ncm_tag.to_ncm_163key(tag_key)
            ncm_163key_xored = bytestrxor(b'c' * len(ncm_163key), ncm_163key)
            ncm_163key_xored_len = len(ncm_163key_xored)
            fd.write(ncm_163key_xored_len.to_bytes(4, 'little'))
            fd.write(ncm_163key_xored)

            fd.write(token_bytes(5))

            cover_data = self.cover_data if self.cover_data else b''
            cover_data_len = len(cover_data)
            fd.write(cover_data_len.to_bytes(4, 'little'))  # cover_space length
            fd.write(cover_data_len.to_bytes(4, 'little'))  # cover_data length
            fd.write(cover_data)

            fd.write(self.getvalue(nocryptlayer=True))

        if filething is None:
            if self.source is None:
                raise TypeError(
                    "attribute 'self.source' and argument 'filething' are empty, "
                    "don't know which file to save to"
                )
            filething = self.source

        if isfilepath(filething):
            with open(filething, mode='wb') as fileobj:
                return operation(fileobj)
        else:
            fileobj = verify_fileobj(filething, 'binary',
                                     verify_seekable=True,
                                     verify_writable=True
                                     )
            return operation(fileobj)

    @classmethod
    def new(cls):
        """返回一个空 NCM 对象。"""
        master_key = (make_random_number_string(29) + make_random_ascii_string(84)).encode('utf-8')

        return cls(ARC4(master_key))

    @property
    def acceptable_ciphers(self) -> list[Type[ARC4]]:
        return [ARC4]

    def __init__(self, cipher: ARC4, /, initial_bytes=b''):
        """基于 BytesIO 的 NCM 透明加密二进制流。

        所有读写相关方法都会经过透明加密层处理：
        读取时，返回解密后的数据；写入时，向缓冲区写入加密后的数据。

        调用读写相关方法时，附加参数 ``nocryptlayer=True``
        可绕过透明加密层，访问缓冲区内的原始加密数据。

        如果你要新建一个 NCM 对象，不要直接调用 ``__init__()``，而是使用构造器方法
        ``NCM.new()`` 和 ``NCM.open()`` 新建或打开已有 NCM 文件，
        使用已有 NCM 对象的 ``save()`` 方法将其保存到文件。

        Args:
            cipher: 要使用的 cipher，必须是一个 libtakiyasha.stdciphers.ARC4 对象
            initial_bytes: 内置缓冲区的初始数据
        """
        super().__init__(cipher, initial_bytes=initial_bytes)

        self._cover_data: bytes | None = None
        self._ncm_tag: CloudMusicIdentifier = CloudMusicIdentifier()
        self._sourcefile: Path | None = None
        self._core_key_deprecated: bytes | None = None

    @property
    def core_key(self) -> bytes | None:
        """（已弃用，且将会在后续版本中删除。）

        核心密钥，用于加/解密主密钥。

        ``NCM.from_file()`` 会在当前对象被创建时设置此属性；而 ``NCM.open()`` 则不会。
        """
        warnings.warn(
            DeprecationWarning(
                f'{type(self).__name__}.core_key is deprecated, no longer used, '
                f'and may be removed in subsequent versions. '
                f'You need to manage the core key by your self.'
            )
        )
        return self._core_key_deprecated

    @core_key.setter
    def core_key(self, value: BytesLike) -> None:
        """（已弃用，且将会在后续版本中删除。）

        核心密钥，用于加/解密主密钥。

        ``NCM.from_file()`` 会在当前对象被创建时设置此属性；而 ``NCM.open()`` 则不会。
        """
        warnings.warn(
            DeprecationWarning(
                f'{type(self).__name__}.core_key is deprecated, no longer used, '
                f'and may be removed in subsequent versions. '
                f'You need to manage the core key by your self.'
            )
        )
        if value is None:
            raise TypeError(
                f"None cannot be assigned to attribute 'core_key'. "
                f"Use `del self.core_key` instead"
            )
        self._core_key_deprecated = tobytes(value)

    @core_key.deleter
    def core_key(self) -> None:
        """（已弃用，且将会在后续版本中删除。）

        核心密钥，用于加/解密主密钥。

        ``NCM.from_file()`` 会在当前对象被创建时设置此属性；而 ``NCM.open()`` 则不会。
        """
        warnings.warn(
            DeprecationWarning(
                f'{type(self).__name__}.core_key is deprecated, no longer used, '
                f'and may be removed in subsequent versions. '
                f'You need to manage the core key by your self.'
            )
        )
        self._core_key_deprecated = None

    @property
    def cover_data(self) -> bytes | None:
        """封面图像数据。"""
        return self._cover_data

    @cover_data.setter
    def cover_data(self, value: BytesLike) -> None:
        """封面图像数据。"""
        if value is None:
            raise TypeError(
                f"None cannot be assigned to attribute 'cover_data'. "
                f"Use `del self.cover_data` instead"
            )
        self._cover_data = tobytes(value)

    @cover_data.deleter
    def cover_data(self) -> None:
        """封面图像数据。"""
        self._cover_data = None

    @property
    def ncm_tag(self) -> CloudMusicIdentifier:
        """163key 的解析结果。"""
        return self._ncm_tag

    @ncm_tag.setter
    def ncm_tag(self, value: CloudMusicIdentifier) -> None:
        """163key 的解析结果。"""
        if not isinstance(value, CloudMusicIdentifier):
            raise TypeError(
                f"attribute 'ncm_tag' must be CloudMusicIdentifier, not {type(value).__name__}"
            )



# -*- coding: utf-8 -*-
from __future__ import annotations

from abc import ABCMeta, abstractmethod
from pathlib import Path
from random import randint
from typing import Callable, Generator, Iterable, Iterator, Literal, Type

try:
    import io
except ImportError:
    import _pyio as io

from .typedefs import IntegerLike, BytesLike, KeyStreamBasedStreamCipherProto, StreamCipherProto, WritableBuffer
from .typeutils import tobytes, toint

__all__ = [
    'CipherSkel',
    'KeyStreamBasedStreamCipherSkel',
    'EncryptedBytesIO'
]


class CipherSkel(metaclass=ABCMeta):
    """适用于一般加密算法的框架类。子类必须实现 ``encrypt()``、``decrypt()``
    和 ``getkey()`` 方法。

    如果以上方法中的任何一个无法实现，应当在被调用时抛出 ``NotImplementedError``。
    """

    @abstractmethod
    def getkey(self, keyname: str = 'master') -> bytes | None:
        raise NotImplementedError

    @abstractmethod
    def encrypt(self, plaindata: BytesLike, /) -> bytes:
        """加密明文 ``plaindata`` 并返回加密结果。

        Args:
            plaindata: 要加密的明文
        """
        raise NotImplementedError

    @abstractmethod
    def decrypt(self, cipherdata: BytesLike, /) -> bytes:
        """解密密文 ``cipherdata`` 并返回解密结果。

        Args:
            cipherdata: 要解密的密文
        """
        raise NotImplementedError


class KeyStreamBasedStreamCipherSkel(metaclass=ABCMeta):
    """适用于简单流式加密算法的框架类。子类必须实现 ``keystream()`` 和 ``getkey()`` 方法。

    以下方法的实现是可选的，但如果实现了，就会被 ``encrypt()`` 和 ``decrypt()`` 使用：

    - ``prexor_encrypt()`` - 加密前对明文的预处理，被 ``encrypt()`` 使用
    - ``postxor_encrypt()`` - 加密后对密文的后处理，被 ``encrypt()`` 使用
    - ``prexor_decrypt()`` - 解密前对密文的预处理，被 ``decrypt()`` 使用
    - ``postxor_decrypt()`` - 解密后对明文的后处理，被 ``decrypt()`` 使用

    以上可选方法的实现必须接受一个类字节对象和一个整数，并返回一个由整数组成的可迭代对象。
    """

    @abstractmethod
    def getkey(self, keyname: str = 'master') -> bytes | None:
        raise NotImplementedError

    @abstractmethod
    def keystream(self,
                  operation: Literal['encrypt', 'decrypt'],
                  nbytes: IntegerLike,
                  offset: IntegerLike, /
                  ) -> Generator[int, None, None]:
        """返回一个生成器对象，对其进行迭代，即可得到从起始点
        ``offset`` 开始，持续一定长度 ``nbytes`` 的密钥流。

        Args:
            operation: 针对特定的操作生成密钥流（加密 encrypt 和解密 decrypt）
            offset: 密钥流的起始点，不应为负数
            nbytes: 密钥流的长度，不应为负数
        """
        raise NotImplementedError

    def encrypt(self, plaindata: BytesLike, offset: IntegerLike = 0, /) -> bytes:
        """加密明文 ``plaindata`` 并返回加密结果。

        Args:
            plaindata: 要加密的明文
            offset: 明文在文件中的位置（偏移量），不应为负数
        """
        plaindata = tobytes(plaindata)
        offset = toint(offset)

        prexor: Callable[[BytesLike, IntegerLike], Iterator[int]] | None = getattr(self, 'prexor_encrypt', None)
        postxor: Callable[[BytesLike, IntegerLike], Iterator[int]] | None = getattr(self, 'postxor_encrypt', None)
        keystream = self.keystream('encrypt', len(plaindata), offset)

        if prexor:
            pd_strm = prexor(plaindata, offset)
        else:
            pd_strm = plaindata
        cd_noxor_strm = (pd_byte ^ ks_byte for pd_byte, ks_byte in zip(pd_strm, keystream))
        if postxor:
            cd_strm = postxor(cd_noxor_strm, offset)
        else:
            cd_strm = cd_noxor_strm

        return bytes(cd_strm)

    def decrypt(self, cipherdata: BytesLike, offset: IntegerLike = 0, /) -> bytes:
        """解密密文 ``cipherdata`` 并返回解密结果。

        Args:
            cipherdata: 要解密的密文
            offset: 密文在文件中的位置（偏移量），不应为负数
        """
        cipherdata = tobytes(cipherdata)
        offset = toint(offset)

        prexor: Callable[[BytesLike, IntegerLike], Iterator[int]] | None = getattr(self, 'prexor_decrypt', None)
        postxor: Callable[[BytesLike, IntegerLike], Iterator[int]] | None = getattr(self, 'postxor_decrypt', None)
        keystream = self.keystream('decrypt', len(cipherdata), offset)

        if prexor:
            cd_strm = prexor(cipherdata, offset)
        else:
            cd_strm = cipherdata
        pd_noxor_strm = (cd_byte ^ ks_byte for cd_byte, ks_byte in zip(cd_strm, keystream))
        if postxor:
            pd_strm = postxor(pd_noxor_strm, offset)
        else:
            pd_strm = pd_noxor_strm

        return bytes(pd_strm)


class EncryptedBytesIO(io.BytesIO):
    def __repr__(self) -> str:
        reprstr_seq = [
            f'<{self.__module__}.{self.__class__.__name__} at {hex(id(self))}, '
            f'cipher {repr(self.cipher)}'
        ]
        if self.source is not None:
            reprstr_seq.append(f", source '{str(self.source)}'")
        reprstr_seq.append('>')

        return ''.join(reprstr_seq)

    @property
    def source(self) -> Path | None:
        """当前对象来源文件的路径。

        在此类的对象中，此属性总是 ``None``。

        如果是通过子类的构造器方法或函数创建的对象，此属性可能会为来源文件的路径，使用 ``Path`` 对象表示。
        """
        if hasattr(self, '_name'):
            return Path(self._name)

    @property
    def cipher(self) -> StreamCipherProto | KeyStreamBasedStreamCipherProto:
        """加解密内置缓冲区数据使用的 Cipher 实例。"""
        return self._cipher

    @property
    def master_key(self) -> bytes | None:
        """主密钥。"""
        return self._cipher.getkey('master')

    @property
    def DEFAULT_BUFFER_SIZE(self) -> int:
        """迭代本对象时，每次迭代返回的数据大小。

        此值必须是一个非零正整数。使用其他值会引发 ``ValueError``。
        """
        return self._DEFAULT_BUFFER_SIZE

    @DEFAULT_BUFFER_SIZE.setter
    def DEFAULT_BUFFER_SIZE(self, value: IntegerLike) -> None:
        bufsize = toint(value)
        if bufsize <= 0:
            raise ValueError(f"attribute 'DEFAULT_BUFFER_SIZE' must greater than 0, got {bufsize}")
        self._DEFAULT_BUFFER_SIZE = bufsize

    @property
    def ITER_METHOD(self) -> Literal['block', 'line']:
        """迭代本对象时使用的迭代模式：按行迭代（``line``）或按固定大小迭代（``block``）。

        仅接受以上两个值，使用其他值会引发 ``ValueError``。
        """
        return self._ITER_METHOD

    @ITER_METHOD.setter
    def ITER_METHOD(self, value: Literal['block', 'line']) -> None:
        if value not in ('block', 'line'):
            raise ValueError(f"attribute 'ITER_METHOD' must be 'block' or 'line', not {repr(value)}")
        self._ITER_METHOD = value

    @property
    def ITER_WITHOUT_CRYPTLAYER(self) -> bool:
        """迭代本对象时，是否直接返回加密数据，而不进行解密。"""
        return self._ITER_WITHOUT_CRYPTLAYER

    @ITER_WITHOUT_CRYPTLAYER.setter
    def ITER_WITHOUT_CRYPTLAYER(self, value: bool) -> None:
        self._ITER_WITHOUT_CRYPTLAYER = bool(value)

    @classmethod
    def verify_stream_cipher(cls,
                             cipher: KeyStreamBasedStreamCipherProto | StreamCipherProto
                             ) -> tuple[bool, bool, bool, bool]:
        keystream_encrypt_available = True
        keystream_decrypt_available = True
        encrypt_available = True
        decrypt_available = True
        # 验证 keystream()
        if isinstance(cipher, KeyStreamBasedStreamCipherProto):
            try:
                ks = cipher.keystream('encrypt', randint(0, 255), randint(0, 8191))
            except NotImplementedError:
                keystream_encrypt_available = False
            except Exception as exc:
                raise TypeError(f"{repr(cipher)} is not a valid cipher object") from exc
            else:
                if not all(map(lambda _: isinstance(_, int), ks)):
                    raise TypeError(
                        f"result of {repr(cipher)}.keystream() returns non-int during iterating"
                    )
            try:
                ks = cipher.keystream('decrypt', randint(0, 255), randint(0, 8191))
            except NotImplementedError:
                keystream_decrypt_available = False
            except Exception as exc:
                raise TypeError(f"{repr(cipher)} is not a valid cipher object") from exc
            else:
                if not all(map(lambda _: isinstance(_, int), ks)):
                    raise TypeError(
                        f"result of {repr(cipher)}.keystream() returns non-int during iterating"
                    )
        elif isinstance(cipher, StreamCipherProto):
            keystream_encrypt_available = False
            keystream_decrypt_available = False
        else:
            raise TypeError(f"{repr(cipher)} is not a cipher object")

        try:
            encrypt_result = cipher.encrypt(bytes([randint(0, 255)]), randint(0, 8191))
        except NotImplementedError:
            encrypt_available = False
        except Exception as exc:
            raise TypeError(f"{repr(cipher)} is not a valid cipher object") from exc
        else:
            if not isinstance(encrypt_result, (bytes, bytearray)):
                raise TypeError(
                    f"{repr(cipher)}.encrypt() returns non-bytes (type {type(encrypt_result).__name__})"
                )

        try:
            decrypt_result = cipher.decrypt(bytes([randint(0, 255)]), randint(0, 8191))
        except NotImplementedError:
            decrypt_available = False
        except Exception as exc:
            raise TypeError(f"{repr(cipher)} is not a valid cipher object") from exc
        else:
            if not isinstance(decrypt_result, (bytes, bytearray)):
                raise TypeError(
                    f"{repr(cipher)}.decrypt() returns non-bytes (type {type(decrypt_result).__name__})"
                )

        return encrypt_available, decrypt_available, keystream_encrypt_available, keystream_decrypt_available

    @property
    def acceptable_ciphers(self) -> list[Type[StreamCipherProto] | Type[KeyStreamBasedStreamCipherProto]]:
        """``__init__()`` 的第一个参数 ``cipher`` 可接受的对象类型。
        在此列表之外的类型会导致 ``__init__()`` 抛出 ``TypeError``。
        """
        return []

    def __init__(self,
                 cipher: StreamCipherProto | KeyStreamBasedStreamCipherProto, /,
                 initial_bytes: BytesLike = b''
                 ) -> None:
        super().__init__(tobytes(initial_bytes))
        self._encrypt_available, self._decrypt_available, self._keystream_encrypt_available, self._keystream_decrypt_available = self.verify_stream_cipher(cipher)
        self._cipher = cipher
        if self.acceptable_ciphers:
            if not isinstance(self._cipher, tuple(self.acceptable_ciphers)):
                if len(self.acceptable_ciphers) == 1:
                    raise TypeError(
                        f"first argument 'cipher' must be {self.acceptable_ciphers[0].__name__}, "
                        f"not {type(self._cipher).__name__}"
                    )
                else:
                    supported_ciphers_str = ', '.join(
                        _.__name__ for _ in self.acceptable_ciphers[:-1]
                    ) + f'or {self.acceptable_ciphers[-1].__name__}'
                    raise TypeError(
                        f"first argument 'cipher' must in {supported_ciphers_str}, "
                        f"not {type(self._cipher).__name__}"
                    )

        self._DEFAULT_BUFFER_SIZE = io.DEFAULT_BUFFER_SIZE
        self._ITER_METHOD: Literal['block', 'line'] = 'block'
        self._ITER_WITHOUT_CRYPTLAYER = False

    def _iterencrypt(self, plaindata: bytes, offset: int, /) -> Generator[int, None, None]:
        if self._keystream_encrypt_available:
            prexor: Callable[[BytesLike, IntegerLike], Iterator[int]] | None = getattr(self._cipher, 'prexor_encrypt', None)
            postxor: Callable[[BytesLike, IntegerLike], Iterator[int]] | None = getattr(self._cipher, 'postxor_encrypt', None)
            keystream = self._cipher.keystream('encrypt', len(plaindata), offset)

            if prexor:
                pd_strm = prexor(plaindata, offset)
            else:
                pd_strm = plaindata
            cd_noxor_strm = (pd_byte ^ ks_byte for pd_byte, ks_byte in zip(pd_strm, keystream))
            if postxor:
                cd_strm = postxor(cd_noxor_strm, offset)
            else:
                cd_strm = cd_noxor_strm
        else:
            cd_strm = self._cipher.encrypt(plaindata, offset)

        yield from cd_strm

    def _iterdecrypt(self, cipherdata: bytes, offset: int, /) -> Generator[int, None, None]:
        if not cipherdata:
            return

        if self._keystream_decrypt_available:
            prexor: Callable[[BytesLike, IntegerLike], Iterator[int]] | None = getattr(self._cipher, 'prexor_decrypt', None)
            postxor: Callable[[BytesLike, IntegerLike], Iterator[int]] | None = getattr(self._cipher, 'postxor_decrypt', None)
            keystream = self._cipher.keystream('decrypt', len(cipherdata), offset)

            if prexor:
                cd_strm = prexor(cipherdata, offset)
            else:
                cd_strm = cipherdata
            pd_noxor_strm = (cd_byte ^ ks_byte for cd_byte, ks_byte in zip(cd_strm, keystream))
            if postxor:
                pd_strm = postxor(pd_noxor_strm, offset)
            else:
                pd_strm = pd_noxor_strm
        else:
            pd_strm = self._cipher.decrypt(cipherdata, offset)

        yield from pd_strm

    def _iterdecrypt_untilnl(self, cipherdata: bytes, offset: int, /) -> Generator[int, None, None]:
        if not cipherdata:
            return

        if self._keystream_decrypt_available:
            prexor: Callable[[BytesLike, IntegerLike], Iterator[int]] | None = getattr(self._cipher, 'prexor_decrypt', None)
            postxor: Callable[[BytesLike, IntegerLike], Iterator[int]] | None = getattr(self._cipher, 'postxor_decrypt', None)
            keystream = self._cipher.keystream('decrypt', len(cipherdata), offset)

            if prexor:
                cd_strm = prexor(cipherdata, offset)
            else:
                cd_strm = cipherdata
            pd_noxor_strm = (cd_byte ^ ks_byte for cd_byte, ks_byte in zip(cd_strm, keystream))
            if postxor:
                pd_strm = postxor(pd_noxor_strm, offset)
            else:
                pd_strm = pd_noxor_strm
        else:
            pd_strm = self._cipher.decrypt(cipherdata, offset)

        for pd_byte in pd_strm:
            yield pd_byte
            if pd_byte == 10:
                return

    def can_encrypt(self) -> bool:
        return self._encrypt_available

    def can_decrypt(self) -> bool:
        return self._decrypt_available

    def getbuffer(self, nocryptlayer: bool = False) -> memoryview:
        if nocryptlayer:
            return super().getbuffer()
        else:
            raise io.UnsupportedOperation('memoryview with crypt layer is not supported')

    def getvalue(self, nocryptlayer: bool = False) -> bytes:
        if nocryptlayer:
            return super().getvalue()
        elif not self.can_decrypt():
            raise io.UnsupportedOperation('getvalue with crypt layer')
        else:
            return bytes(self._iterdecrypt(super().getvalue(), 0))

    def read(self, size: IntegerLike | None = -1, /, nocryptlayer: bool = False) -> bytes:
        if nocryptlayer:
            return super().read(size)
        elif not self.can_decrypt():
            raise io.UnsupportedOperation('read with crypt layer')
        else:
            curpos = self.tell()
            if size is None:
                size = -1
            else:
                size = toint(size)
            if size < 0:
                target_data = super().getvalue()[curpos:]
            else:
                target_data = super().getvalue()[curpos:curpos + size]

            result = bytes(self._iterdecrypt(target_data, curpos))
            self.seek(curpos + len(result), 0)

            return result

    def read1(self, size: IntegerLike | None = -1, /, nocryptlayer: bool = False) -> bytes:
        if not (nocryptlayer or self.can_decrypt()):
            raise io.UnsupportedOperation('read1 with crypt layer')
        return self.read(size, nocryptlayer)

    def readline(self, size: IntegerLike | None = -1, /, nocryptlayer: bool = False) -> bytes:
        if nocryptlayer:
            return super().readline(size)
        elif not self.can_decrypt():
            raise io.UnsupportedOperation('readline with crypt layer')
        else:
            curpos = self.tell()
            if size is None:
                size = -1
            else:
                size = toint(size)
            if size < 0:
                blksize = self._DEFAULT_BUFFER_SIZE
                results = []
                while 1:
                    target_data = super().getvalue()[curpos:curpos + blksize]
                    if not target_data:
                        return b''
                    result_data = bytes(self._iterdecrypt_untilnl(target_data, curpos))
                    self.seek(curpos + len(result_data), 0)
                    curpos += len(result_data)
                    results.append(result_data)
                    if len(result_data) != len(target_data):
                        return b''.join(results)
            else:
                target_data = super().getvalue()[curpos:curpos + size]

                result = bytes(
                    self._iterdecrypt_untilnl(target_data, curpos)
                )
                self.seek(curpos + len(result), 0)

            return result

    def readlines(self,
                  hint: IntegerLike | None = None, /,
                  nocryptlayer: bool = False
                  ) -> list[bytes]:
        if nocryptlayer:
            return super().readlines(hint)
        elif not self.can_decrypt():
            raise io.UnsupportedOperation('readlines with crypt layer')
        else:
            curpos = self.tell()
            max_read_size = len(super().getvalue()[curpos:])
            if hint is None:
                hint = -1
            else:
                hint = toint(hint)
            if hint < 0:
                read_size = max_read_size
            else:
                read_size = min(hint, max_read_size)

            nbytes = 0
            lines = []
            while 1:
                line = self.readline()
                lines.append(line)
                nbytes += len(line)
                if nbytes >= read_size:
                    break

            return lines

    def readinto(self, buffer: WritableBuffer, /, nocryptlayer: bool = False) -> int:
        if nocryptlayer:
            return super().readinto(buffer)
        elif not self.can_decrypt():
            raise io.UnsupportedOperation('readinto with crypt layer')
        else:
            if not isinstance(buffer, memoryview):
                buffer = memoryview(buffer)
            buffer = buffer.cast('B')

            data = self.read(len(buffer))
            data_len = len(data)

            buffer[:data_len] = data

            return data_len

    def readinto1(self, buffer: WritableBuffer, /, nocryptlayer: bool = False) -> int:
        if not (nocryptlayer or self.can_decrypt()):
            raise io.UnsupportedOperation('readinto1 with crypt layer')
        return self.readinto(buffer, nocryptlayer)

    def write(self, data: BytesLike, /, nocryptlayer: bool = False) -> int:
        if nocryptlayer:
            return super().write(data)
        elif not self.can_encrypt():
            raise io.UnsupportedOperation('write with crypt layer')
        else:
            curpos = self.tell()
            return super().write(
                bytes(
                    self._iterencrypt(tobytes(data), curpos)
                )
            )

    def writelines(self, lines: Iterable[BytesLike], /, nocryptlayer: bool = False) -> None:
        if not (nocryptlayer or self.can_encrypt()):
            raise io.UnsupportedOperation('write with crypt layer')
        for data in lines:
            self.write(data, nocryptlayer)

    def __iter__(self):
        return self

    def __next__(self) -> bytes:
        if self._ITER_METHOD == 'line':
            ret = self.readline(nocryptlayer=self._ITER_WITHOUT_CRYPTLAYER)
            if not ret:
                raise StopIteration
            return ret
        elif self._ITER_METHOD == 'block':
            ret = self.read(self._DEFAULT_BUFFER_SIZE, nocryptlayer=self._ITER_WITHOUT_CRYPTLAYER)
            if not ret:
                raise StopIteration
            return ret
        else:
            raise ValueError(
                f"attribute 'ITER_METHOD' must be 'block' or 'line', not {repr(self._ITER_METHOD)}"
            )



# -*- coding: utf-8 -*-
from __future__ import annotations

import re
from functools import lru_cache
from pathlib import Path
from typing import Literal, NamedTuple

__all__ = ['version', 'version_info', 'progname']

_VERSION_PATTERN_STR = r"""
    v?
    (?:
        (?:(?P<epoch>[0-9]+)!)?                           # epoch
        (?P<release>[0-9]+(?:\.[0-9]+)*)                  # release segment
        (?P<pre>                                          # pre-release
            [-_\.]?
            (?P<pre_l>(a|b|c|rc|alpha|beta|pre|preview))
            [-_\.]?
            (?P<pre_n>[0-9]+)?
        )?
        (?P<post>                                         # post release
            (?:-(?P<post_n1>[0-9]+))
            |
            (?:
                [-_\.]?
                (?P<post_l>post|rev|r)
                [-_\.]?
                (?P<post_n2>[0-9]+)?
            )
        )?
        (?P<dev>                                          # dev release
            [-_\.]?
            (?P<dev_l>dev)
            [-_\.]?
            (?P<dev_n>[0-9]+)?
        )?
    )
    (?:\+(?P<local>[a-z0-9]+(?:[-_\.][a-z0-9]+)*))?       # local version
"""
_VERSION_PATTERN = re.compile(r"^\s*" + _VERSION_PATTERN_STR + r"\s*$", re.VERBOSE | re.IGNORECASE)


class _VersionInfo(NamedTuple):
    major: int
    minor: int = None
    micro: int = None
    epoch: int | None = None
    is_pre: bool = False
    pre_type: Literal['alpha', 'beta', 'rc'] | None = None
    pre_num: int | None = None
    is_post: bool = False
    post_num1: int | None = None
    post_num2: int | None = None
    is_dev: bool = False
    dev_num: int | None = None
    local_identifier: str | None = None


@lru_cache
def progname() -> str:
    ret = __name__.split('.')[0]
    if ret == '__main__':
        for parentpath in Path(__file__).parents:
            if (parentpath / '__init__.py').exists():
                if (parentpath / '__main__.py').exists():
                    if (parentpath / 'VERSION').exists():
                        return parentpath.name
        raise RuntimeError('cannot get the program name')
    return ret


@lru_cache
def version() -> str:
    for parentpath in Path(__file__).parents:
        if parentpath.name == progname():
            if (parentpath / '__init__.py').exists():
                if (parentpath / 'VERSION').exists():
                    with open(parentpath / 'VERSION', encoding='utf-8') as version_file:
                        return version_file.readline().strip()
    raise RuntimeError('cannot get the version')


@lru_cache
def version_info() -> _VersionInfo:
    result = _VERSION_PATTERN.search(version()).groupdict()
    params: dict[str, str | int | bool | None] = {}
    for k, v in result.items():
        if k == 'epoch':
            if v is not None:
                params['epoch'] = int(v)
        elif k == 'release':
            rel = [int(_) for _ in v.split('.', maxsplit=3)]
            rel += [0] * (3 - len(rel))
            params['major'] = int(rel[0])
            params['minor'] = int(rel[1])
            params['micro'] = int(rel[2])
        elif k == 'pre':
            if v is not None:
                params['is_pre'] = True
        elif k == 'pre_l' and params.get('is_pre', False):
            if v == 'a':
                params['pre_type'] = 'alpha'
            elif v == 'b':
                params['pre_type'] = 'beta'
            elif v in ('c', 'rc'):
                params['pre_type'] = 'rc'
        elif k == 'pre_n' and params.get('is_pre', False):
            if v is not None:
                params['pre_num'] = int(v)
        elif k == 'post':
            if v is not None:
                params['is_post'] = True
        elif k == 'post_n1' and params.get('is_post', False):
            if v is not None:
                params['post_num1'] = int(v)
        elif k == 'post_n2' and params.get('is_post', False):
            if v is not None:
                params['post_num2'] = int(v)
        elif k == 'dev':
            if v is not None:
                params['is_dev'] = True
        elif k == 'dev_n' and params.get('is_dev', False):
            if v is not None:
                params['dev_num'] = int(v)
        elif k == 'local':
            params['local_identifier'] = v
            break

    return _VersionInfo(**params)



#!/bin/env python
# -*- coding: utf-8 -*-
from __future__ import annotations


class LibTakiyashaWarning(Warning):
    pass


class CipherEncryptingWarning(LibTakiyashaWarning):
    pass


class CipherDecryptingWarning(LibTakiyashaWarning):
    pass


class CrypterCreatingWarning(LibTakiyashaWarning):
    pass


class CrypterSavingWarning(LibTakiyashaWarning):
    pass
